"""SciPy solver Class."""

from typing import Callable, Union

import numpy as np
from numpy.typing import ArrayLike
from scipy.optimize import minimize, Bounds

from ._base_solvers import NonLinearOptimizer


class OptimizerSciPy(NonLinearOptimizer):
    """
    SciPy solver configuration class.

    Attributes
    ----------
    options : dict
        Options for the solver
    method : Union[str, Callable]
        The solver method
    """

    name = "SciPy minimize"
    short_name = "scipy"

    options: dict[str]
    method: Union[str, Callable]

    def __init__(self):
        self.options = {
            "disp": False,
            "ftol": 1e-4,
        }

        self.method = "L-BFGS-B"

        super().__init__()

    def solve(self, x0: ArrayLike) -> tuple[np.ndarray, dict]:
        """
        Solve the problem.

        Parameters
        ----------
        x0 : np.ndarray
            Initial guess for the decision variables.

        Returns
        -------
        result : dict
        """

        self.options.update({"maxiter": self.max_iter})

        x0 = np.asarray(x0)

        bounds = Bounds(lb=self.bounds[0], ub=self.bounds[1])

        # Initialize the SciPy solution function and its arguments
        result = minimize(
            x0=x0,
            fun=self.objective,
            method=self.method,
            jac=self.gradient,
            # constraints=self.constraints,
            # hess=self.hessian,
            tol=self.abs_obj_tol,
            bounds=bounds,
            options=self.options,
        )

        return result["x"], result



import math
from typing import Union
import numpy as np
from numpy.typing import ArrayLike
from tqdm import tqdm

from pyRadPlan.geometry import lps
from pyRadPlan.plan import Plan
from ._base import StfGeneratorBase


class StfGeneratorExternalBeam(StfGeneratorBase):
    """
    Base class for geometry enerated for external beam therapy.

    Provides the basic functionality and infrastructure for external beam
    therapy, such as the definition of gantry / couch angles, an isocenter,
    and basic ray geometry transformations.

    Parameters
    ----------
    pln : Plan, optional
        A Plan object. If given, configuration will be loaded from the plan.

    Attributes
    ----------
    gantry_angles : list[float]
        A list of gantry angles.
    couch_angles : list[float]
        A list of couch angles.
    iso_center : Union[ArrayLike, None]
        The isocenter coordinates.
    """

    gantry_angles: list[float]
    couch_angles: list[float]
    iso_center: Union[ArrayLike, None]

    @property
    def num_of_beams(self) -> int:
        """Number of beams."""
        if len(self.gantry_angles) != len(self.couch_angles):
            raise ValueError("Inconsistent number of gantry and couch angles.")

        return len(self.gantry_angles)

    def __init__(self, pln: Plan = None):
        self.gantry_angles = [0.0]
        self.couch_angles = [0.0]
        self.iso_center = None

        super().__init__(pln)


class StfGeneratorExternalBeamRayBixel(StfGeneratorExternalBeam):
    """
    Base class for ray-bixel-based geometry for IMRT.

    Attributes
    ----------
    bixel_width : float
        The bixel width (beamlet / spot spacing).
    """

    bixel_width: float

    def __init__(self, pln: Plan = None):
        self.bixel_width = 5.0

        super().__init__(pln)

    def _computed_target_margin(self) -> float:
        """Target margin to apply for beamlet placing."""
        return self.bixel_width

    def _create_rays(self, beam) -> list[dict]:
        """Get the rays on a beam / stf element."""

        ray_pos = self._generate_ray_positions_in_isocenter_plane(beam)

        # Get the rotation matrix in LPS rotating the coordinates into the BEV System
        rotation_matrix = lps.get_beam_rotation_matrix(beam["gantry_angle"], beam["couch_angle"])

        rays = list[dict]()
        for j in range(ray_pos.shape[1]):
            ray = {
                "ray_pos_bev": ray_pos[:, j].flatten(),
                "ray_pos": rotation_matrix @ ray_pos[:, j].flatten(),
            }
            rays.append(ray)

        return rays

    def _generate_ray_positions_in_isocenter_plane(self, beam) -> np.ndarray:
        """Generate the ray positions for the STF in the isocenter plane."""

        rotation_matrix_t = lps.get_beam_rotation_matrix(
            beam["gantry_angle"], beam["couch_angle"]
        ).transpose()

        sad = beam["sad"]
        bw = beam["bixel_width"]

        iso_coords = self._target_voxel_coordinates - beam["iso_center"].reshape((3, 1))

        # Get target voxel coordinates in BEV in the isocenter plane
        bev_coords_isoplane = rotation_matrix_t @ iso_coords
        bev_coords_isoplane = sad * bev_coords_isoplane / (sad + bev_coords_isoplane[1, :])
        bev_coords_isoplane[1, :] = 0.0

        # Obtain unique ray positions
        ray_pos = np.unique(bw * np.round(bev_coords_isoplane / bw), axis=1)

        # Obtain unique ray positions
        # Pad ray position array if resolution of target voxel grid not sufficient
        max_ct_resolution = max(
            [self._ct.resolution["x"], self._ct.resolution["y"], self._ct.resolution["z"]]
        )
        if self.bixel_width < max_ct_resolution:
            orig_ray_pos = ray_pos.copy()
            for j in range(
                -math.floor(max_ct_resolution / self.bixel_width),
                math.floor(max_ct_resolution / self.bixel_width),
            ):
                for k in range(
                    -math.floor(max_ct_resolution / self.bixel_width),
                    math.floor(max_ct_resolution / self.bixel_width),
                ):
                    if abs(j) + abs(k) == 0:
                        continue
                    ray_pos = np.hstack(
                        (
                            ray_pos,
                            orig_ray_pos
                            + np.array((j * self.bixel_width, 0.0, k * self.bixel_width)).reshape(
                                (3, 1)
                            ),
                        )
                    )

        # TODO: for DAO, there is a filling of empty bixels in matRad

        ray_pos = np.unique(ray_pos, axis=1)

        return ray_pos

    def _generate_source_geometry(self):
        """Generate the source geometry for the STF."""

        # Validate bixel width
        if self.bixel_width <= 0.0:
            raise ValueError(
                "Bixel width (spot distance) needs to be a real number [mm] larger than zero."
            )

        # Validate iso center
        if self.iso_center is None:
            self.iso_center = self._cst.target_center_of_mass().reshape((1, 3))
        else:
            self.iso_center = np.asarray(self.iso_center, dtype=np.float64).reshape((-1, 3))

        # Now check isocenter
        if (
            self.iso_center.shape[0] != self.num_of_beams or self.iso_center.shape[0] != 1
        ) and self.iso_center.shape[1] != 3:
            raise ValueError(
                "Iso center needs to have three coordinates (1x3 array) ",
                " or a set of 3D coordinates for each beam (nx3 array).",
            )

        if self.iso_center.shape[0] == 1:
            self.iso_center = np.repeat(self.iso_center, self.num_of_beams, axis=0)

        stf = []

        for i in tqdm(range(self.num_of_beams), desc="Beam", unit="b", leave=False):
            beam = {}

            # Correct for iso center position. With this correction isocenter is (0, 0, 0) [mm]
            # coordsX = (self.coordsX_vox * self.ct.resolution["x"] - self.iso_center[i][0])
            # coordsY = (self.coordsY_vox * self.ct.resolution["y"] - self.iso_center[i][1])
            # coordsZ = (self.coordsZ_vox * self.ct.resolution["z"] - self.iso_center[i][2])

            # Save meta information for treatment plan
            beam["gantry_angle"] = self.gantry_angles[i]
            beam["couch_angle"] = self.couch_angles[i]
            beam["bixel_width"] = self.bixel_width
            beam["radiation_mode"] = self.radiation_mode
            beam["sad"] = self.machine.sad
            beam["iso_center"] = self.iso_center[i]
            beam["machine"] = self.machine.name
            beam["source_point_bev"] = np.array([0.0, -beam["sad"], 0.0], dtype=float)

            # Get the rotation matrix in LPS rotating the coordinates into the BEV system
            rotation_matrix = lps.get_beam_rotation_matrix(
                beam["gantry_angle"], beam["couch_angle"]
            )

            beam["source_point"] = rotation_matrix @ beam["source_point_bev"]

            beam["rays"] = self._create_rays(beam)

            stf.append(beam)
        return stf



import logging

import numpy as np

from pyRadPlan import (
    IonPlan,
    generate_stf,
    calc_dose_influence,
    fluence_optimization,
    plot_slice,
    load_tg119,
)

from pyRadPlan.optimization.objectives import SquaredDeviation, SquaredOverdosing, MeanDose

logging.basicConfig(level=logging.INFO)

# Load TG119 (provided within pyRadPlan)
ct, cst = load_tg119()

# Create a plan object
pln = IonPlan(radiation_mode="protons", machine="Generic")
pln.prop_opt = {"solver": "scipy"}

# Generate Steering Geometry ("stf")
stf = generate_stf(ct, cst, pln)

# Calculate Dose Influence Matrix ("dij")
dij = calc_dose_influence(ct, cst, stf, pln)

# Optimization
cst.vois[1].objectives = [SquaredDeviation(priority=100.0, d_ref=3.0)]  # Target
cst.vois[0].objectives = [SquaredOverdosing(priority=10.0, d_max=1.0)]  # OAR
cst.vois[2].objectives = [
    MeanDose(priority=1.0, d_ref=0.0),
    SquaredOverdosing(priority=10.0, d_max=2.0),
]  # BODY
fluence = fluence_optimization(ct, cst, stf, dij, pln)

# Result
result = dij.compute_result_ct_grid(fluence)

# Choose a slice to visualize
view_slice = int(np.round(ct.size[2] / 2))

# Visualize
plot_slice(
    ct=ct,
    cst=cst,
    overlay=result["physical_dose"],
    view_slice=view_slice,
    plane="axial",
    overlay_unit="Gy",
)



"""Squared Overdosing Objective."""

from typing import Annotated
from pydantic import Field

from numba import njit
from numpy import clip

from ._objective import Objective, ParameterMetadata

# %% Class definition


class SquaredOverdosing(Objective):
    """
    Squared Overdosing (piece-wise positive least-squares) objective.

    Attributes
    ----------
    d_max : float
        maximum dose value (above which we penalize)
    """

    name = "Squared Overdosing"

    d_max: Annotated[float, Field(default=30.0, ge=0.0), ParameterMetadata(kind="reference")]

    def compute_objective(self, values):
        return _compute_objective(values, self.d_max)

    def compute_gradient(self, values):
        return _compute_gradient(values, self.d_max)


@njit
def _compute_objective(dose, d_max):
    overdose = clip(dose - d_max, a_min=0, a_max=None)

    return (overdose @ overdose) / len(dose)


@njit
def _compute_gradient(dose, d_max):
    overdose = clip(dose - d_max, a_min=0, a_max=None)
    return 2.0 * overdose / len(overdose)



import pytest

try:
    import matlab
except ImportError:
    pytest.skip("Matlab not installed", allow_module_level=True)


@pytest.fixture
def engine():
    from pyRadPlan.matRad import MatRadEngineMatlab

    return MatRadEngineMatlab()


def test_engineBasicCall(engine):
    basicZeros = engine.zeros(10)
    assert basicZeros.size == (10, 10)  # Matlab returns matlab specific array (matlab.double)


def test_engineParameters(engine):
    assert isinstance(engine.matRadVersionStr, str)
    assert isinstance(engine.engineVersionStr, str)
    assert isinstance(engine.matRadPath, str)


def test_engineInitMatrad(engine):
    engine.matRad_rc



"""Siddon Ray Tracing Algorithm for Voxelized Geometry."""

from typing import Union
import time
import logging

import numpy as np
import SimpleITK as sitk

from pyRadPlan.raytracer._base import RayTracerBase

# from ._perf import _fast_compute_all_alphas, _fast_compute_plane_alphas
logger = logging.getLogger(__name__)


class RayTracerSiddon(RayTracerBase):
    """Siddon Ray Tracing Algorithm through voxelized geometry."""

    debug_core_performance: bool

    def __init__(self, cubes: Union[sitk.Image, list[sitk.Image]]):
        self.debug_core_performance = False
        super().__init__(cubes)

    # @jit(nopython=True)
    def trace_ray(
        self,
        isocenter: Union[list, np.ndarray],
        source_points: Union[list, np.ndarray],
        target_points: Union[list, np.ndarray],
    ) -> tuple[np.ndarray, np.ndarray, list[np.ndarray], np.ndarray, np.ndarray]:
        """Trace an individual ray."""

        # Check type and dimension of target_point (should be nx3)
        # Check type and dimension of source_point (should be nx3)
        # Check type and dimension of cubes (should be either a list of or a single 3D numpy array)
        # Check type and dimension of resolution (should be a dictionary with keys 'x', 'y', 'z'
        # and values as floats)
        isocenter = np.asarray(isocenter)
        source_points = np.asarray(source_points)
        target_points = np.asarray(target_points)

        if target_points.size != 3 or source_points.size != 3 or isocenter.size != 3:
            raise ValueError(
                "Number of target Points and source points needs to be equal to one! If you want "
                "to trace multiple rays at once, use trace_rays instead!"
            )
        alphas, lengths, rho, d12, ix = self.trace_rays(
            isocenter, source_points.reshape((1, 3)), target_points.reshape((1, 3))
        )

        # Squeeze Dimensions
        alphas = alphas.squeeze()
        lengths = lengths.squeeze()
        rho = [r.squeeze() for r in rho]
        ix = ix.squeeze()

        return alphas, lengths, rho, d12, ix

    def trace_rays(
        self,
        isocenter: Union[list, np.ndarray],
        source_points: Union[list, np.ndarray],
        target_points: Union[list, np.ndarray],
    ) -> tuple[np.ndarray, np.ndarray, list[np.ndarray], np.ndarray, np.ndarray]:
        """
        Vectorized Implementation of RayTracing.

        Uses padding to create matrices of ray information.

        Notes
        -----
        Currently, the vectorized implementation uses padding with NaN values. This is not the most
        efficient way to handle the different lengths of the rays. A more efficient way would be to
        use more performant padding values (e.g. an unrealistically large value like the respective
        maximum floating point value)
        """

        num_rays = target_points.shape[0]
        num_sources = source_points.shape[0]

        if num_sources not in (num_rays, 1):
            raise ValueError(
                f"Number of source points ({num_sources}) needs to be one or equal to number of "
                f"target points ({num_rays})!"
            )
        if num_sources == 1:
            source_points = np.tile(source_points, (num_rays, 1))
            num_sources = num_rays

        self._source_points = (source_points + isocenter).astype(self.precision)
        self._target_points = (target_points + isocenter).astype(self.precision)
        self._ray_vec = self._target_points - self._source_points

        t_allalphas_start = time.perf_counter()
        alphas = self._compute_all_alphas()
        t_allalphas_end = time.perf_counter()
        t_allalphas_elapsed = t_allalphas_end - t_allalphas_start

        d12 = np.linalg.norm(self._ray_vec, axis=1, keepdims=True)
        tmp_diff = np.diff(alphas, axis=1)

        lengths = d12 * tmp_diff
        alphas_mid = alphas[:, :-1] + 0.5 * tmp_diff

        val_ix, ijk = self._compute_indices_from_alpha(alphas_mid)

        t_indices_end = time.perf_counter()
        t_indices_elapsed = t_indices_end - t_allalphas_end

        valid_indices = np.nonzero(val_ix)

        if all(arr.size == 0 for arr in valid_indices):
            alphas = np.empty((num_rays, 0), dtype=self.precision)
            lengths = np.empty((num_rays, 0), dtype=self.precision)
            rho = [np.empty((num_rays, 0), dtype=self.precision) for _ in self._cubes]
            ix = np.empty((num_rays, 0), dtype=np.int64)

        else:
            rho, ix = self._get_rho_and_indices(val_ix, valid_indices, ijk)

        t_finalization_end = time.perf_counter()
        t_finalization_elapsed = t_finalization_end - t_indices_end
        if self.debug_core_performance:
            logger.debug(
                f"Trace Ray: {num_rays} rays, {num_sources} sources, "
                f"compute_all_alphas: {t_allalphas_elapsed:.4f}s, "
                f"compute_indices: {t_indices_elapsed:.4f}s, "
                f"finalization: {t_finalization_elapsed:.4f}s"
            )

        return alphas, lengths, rho, d12, ix

    def _compute_all_alphas(self) -> np.ndarray:
        """
        Compute all rays' alpha values (length to plane intersections).

        Here we setup grids to enable logical indexing when computing
        the alphas along each dimension. All alphas between the
        minimum and maximum index will be computed, with additional
        exclusion of singular plane occurrences (max == min)
        All values out of scope will be set to NaN.
        """

        alpha_limits = self._compute_alpha_limits()

        i_min, i_max, j_min, j_max, k_min, k_max = self._compute_entry_and_exit(alpha_limits)

        # Compute alphas for each plane and merge parametric sets

        alpha_x = self._compute_plane_alphas(
            i_min,
            i_max,
            self._x_planes,
            self._source_points[:, 0],
            self._ray_vec[:, 0],
        )
        alpha_y = self._compute_plane_alphas(
            j_min,
            j_max,
            self._y_planes,
            self._source_points[:, 1],
            self._ray_vec[:, 1],
        )
        alpha_z = self._compute_plane_alphas(
            k_min,
            k_max,
            self._z_planes,
            self._source_points[:, 2],
            self._ray_vec[:, 2],
        )

        alphas = np.concatenate((alpha_limits, alpha_x, alpha_y, alpha_z), axis=1)

        # Vectorized unique operation across rows
        # Sort alphas row-wise and remove duplicates
        alphas.sort(axis=1)  # Sort each row ascendingly
        mask = np.diff(alphas, axis=1) == 0  # Identify duplicates
        alphas[:, 1:][mask] = np.nan  # Replace duplicates with NaN
        alphas.sort(axis=1)

        # Alternative for loop to the last sorting operation in O(n)
        # Could be numba'd to squeeze out more performance
        # for row in alphas:
        #     mask = ~np.isnan(row)
        #     k = np.sum(mask)  # how many non-NaN in this row
        #     row[:k] = row[mask]      # move them to the front
        #     row[k:] = np.nan         # fill the rest with NaN

        # Size Reduction
        max_num_columns = np.max(np.sum(~np.isnan(alphas), axis=1))
        alphas = alphas[:, :max_num_columns]

        return alphas

    def _compute_plane_alphas(
        self,
        dim_min: np.ndarray,
        dim_max: np.ndarray,
        planes: np.ndarray,
        source: np.ndarray,
        ray: np.ndarray,
    ) -> np.ndarray:
        """
        Compute the alphas for a given plane.

        Parameters
        ----------
        dim_min : np.ndarray
            The minimum dimension of the plane.
        dim_max : np.ndarray
            The maximum dimension of the plane.
        planes : np.ndarray
            The planes to compute the alphas for.
        source : np.ndarray
            The source points.
        ray : np.ndarray
            The ray vectors.

        Returns
        -------
        alphas : np.ndarray
            The computed alphas for the given plane.
        """

        # ensure 1-D

        # 1) make a (1, P) index row, and compare to (N, 1) dim_min/max → (N, P) mask
        plane_ix = np.arange(planes.shape[0])[None, :]  # shape (1, P)
        low = plane_ix < dim_min[:, None]  # before entry
        high = plane_ix > dim_max[:, None]  # after exit
        deg = (plane_ix == dim_min[:, None]) & (plane_ix == dim_max[:, None])
        nanm = np.isnan(dim_min)[:, None] | np.isnan(dim_max)[:, None]

        mask_invalid = low | high | deg | nanm  # shape (N, P)

        # 2) compute all alphas in one shot (broadcasted): (planes[None,:] - source[:,None]) / ray[:,None]
        #    guard against divide-by-zero warnings if you like via errstate or inv_ray trick
        with np.errstate(divide="ignore", invalid="ignore"):
            alphas = (planes[None, :] - source[:, None]) / ray[:, None]

        # 3) mask out invalid entries
        alphas[mask_invalid] = np.nan

        return alphas

    def _compute_alpha_limits(self):
        """
        Compute the alpha limits for the ray tracing.

        This is a helper function to compute the alpha limits for the ray tracing.
        It is used in the trace_rays function to compute the alpha limits for each ray.
        """

        # Draft for faster alpha calculation
        # # plane limits
        p_min = np.asarray(
            [self._x_planes[0], self._y_planes[0], self._z_planes[0]], dtype=self.precision
        )
        p_max = np.asarray(
            [self._x_planes[-1], self._y_planes[-1], self._z_planes[-1]], dtype=self.precision
        )

        # 1) raw alpha to the two planes per axis, shape (N, 3, 2)

        with np.errstate(divide="ignore", invalid="ignore"):
            alpha_planes = np.stack(
                (
                    (p_min - self._source_points) / self._ray_vec,  # alpha to "near" plane
                    (p_max - self._source_points) / self._ray_vec,
                ),  # alpha to "far"  plane
                axis=-1,
            )

        # 2) final [alpha_min, alpha_max] per axis with ray_vec != 0 mask
        alpha_axis_min = np.nanmin(alpha_planes, axis=-1)  # (N, 3)
        alpha_axis_max = np.nanmax(alpha_planes, axis=-1)  # (N, 3)

        zero_mask = (self._ray_vec == 0.0).all(axis=1)  # (N,)
        alpha_axis_min[zero_mask] = 0.0
        alpha_axis_max[zero_mask] = 1.0

        alpha_min_values = np.maximum(0.0, np.nanmax(alpha_axis_min, axis=1))  # (N,)
        alpha_max_values = np.minimum(1.0, np.nanmin(alpha_axis_max, axis=1))  # (N,)

        alpha_limits = np.stack((alpha_min_values, alpha_max_values), axis=1)  # (N, 2)

        return alpha_limits

    def _compute_indices_from_alpha(self, alphas_mid: np.ndarray):
        cube_origin = np.asarray(self._cubes[0].GetOrigin())

        # Compute coordinates
        sp_scaled = (self._source_points - cube_origin) / self._resolution
        rv_scaled = self._ray_vec / self._resolution

        ijk = sp_scaled[:, :, None] + rv_scaled[:, :, None] * alphas_mid[:, None, :]
        ijk[~np.isfinite(ijk)] = -1.0

        # Round in place
        np.round(ijk, out=ijk)
        ijk = ijk.astype(np.int32, copy=False)

        cube_dim_brd = self._cube_dim[None, :, None]
        val_ix = ((ijk >= 0) & (ijk < cube_dim_brd)).all(axis=1)

        return val_ix, ijk

    def _get_rho_and_indices(self, val_ix: np.ndarray, valid_indices: np.ndarray, ijk: np.ndarray):
        """
        Finalize the output of densities and indices.

        Returns
        -------
        rho : list[np.ndarray]
            The rho values for each cube.
        ix : np.ndarray
            The indices within the cubes.
        """
        stride_j = self._cube_dim[2]
        stride_i = self._cube_dim[1] * self._cube_dim[2]

        ix = ijk[:, 2, :].astype(np.int64, copy=True)
        ix += stride_j * ijk[:, 1, :]
        ix += stride_i * ijk[:, 0, :]

        ix[~val_ix] = -1

        rho = [np.full(val_ix.shape, np.nan, dtype=self.precision) for _ in self._cubes]
        for s, cube in enumerate(self._cubes):
            # Views SimpleITKs image buffer as a numpy array, preserving dimension ordering of
            # sitk
            cube_linear = sitk.GetArrayViewFromImage(cube).ravel(order="F")
            rho[s][valid_indices] = cube_linear[ix[valid_indices]]

        return rho, ix

    def _compute_entry_and_exit(self, alpha_limits: np.ndarray):
        """
        Compute the entry and exit points for the ray tracing.

        This is a helper function to compute the entry and exit points for the ray tracing.
        It is used in the trace_rays function to compute the entry and exit points for each ray.
        """

        ray_direction_positive = self._ray_vec > 0
        alpha_limits_reverse = alpha_limits[:, ::-1]

        alpha_axis = np.where(
            ray_direction_positive[:, :, None],
            alpha_limits[:, None, :],
            alpha_limits_reverse[:, None, :],
        )

        lower_planes = np.array(
            [self._x_planes[0], self._y_planes[0], self._z_planes[0]], dtype=self.precision
        )
        upper_planes = np.array(
            [self._x_planes[-1], self._y_planes[-1], self._z_planes[-1]], dtype=self.precision
        )

        nplanes = np.asarray(self._num_planes, dtype=self.precision)

        dim_min = (
            nplanes[None, :]
            - (upper_planes - alpha_axis[:, :, 0] * self._ray_vec - self._source_points)
            / self._resolution[None, :]
            - 1
        )
        dim_max = (
            self._source_points + alpha_axis[:, :, 1] * self._ray_vec - lower_planes
        ) / self._resolution[None, :]

        # Rounding
        dim_min = np.ceil(np.round(1000 * dim_min) / 1000)
        dim_max = np.floor(np.round(1000 * dim_max) / 1000)

        # unpack the dimensions to i, j, k
        i_min, j_min, k_min = dim_min.T
        i_max, j_max, k_max = dim_max.T

        return i_min, i_max, j_min, j_max, k_min, k_max

    def _initialize_geometry(self):
        """
        Initialize the geometry for the ray tracing.

        Notes
        -----
        For a detailed description of the variables, see Siddon 1985 Medical Physics.
        """

        ref_cube = self._cubes[0]

        if ref_cube.GetDimension() != 3:
            raise ValueError("Only 3D cubes are supported by RayTracerSiddon!")

        origin = np.asarray(ref_cube.GetOrigin()).astype(self.precision)
        self._resolution = np.asarray(ref_cube.GetSpacing()).astype(self.precision)
        direction = np.asarray(ref_cube.GetDirection()).reshape(3, 3).astype(self.precision)
        self._cube_dim = np.asarray(ref_cube.GetSize())

        increment = np.zeros_like(origin)
        increment[0] = (direction @ np.array([1, 0, 0], dtype=self.precision))[
            0
        ] * self._resolution[0]
        increment[1] = (direction @ np.array([0, 1, 0], dtype=self.precision))[
            1
        ] * self._resolution[1]
        increment[2] = (direction @ np.array([0, 0, 1], dtype=self.precision))[
            2
        ] * self._resolution[2]

        self._x_planes = (
            origin[0]
            + (np.arange(self._cube_dim[0] + 1, dtype=self.precision) - 0.5) * increment[0]
        )
        self._y_planes = (
            origin[1]
            + (np.arange(self._cube_dim[1] + 1, dtype=self.precision) - 0.5) * increment[1]
        )
        self._z_planes = (
            origin[2]
            + (np.arange(self._cube_dim[2] + 1, dtype=self.precision) - 0.5) * increment[2]
        )

        self._num_planes = [len(self._x_planes), len(self._y_planes), len(self._z_planes)]



import pytest
import os
import sys

if sys.version_info < (3, 10):
    import importlib_resources as resources  # Backport for older versions
else:
    from importlib import resources  # Standard from Python 3.9+

from pyRadPlan.io._matlab_file_handler import MatlabFileHandler
import pyRadPlan.io.matfile as matfile
from pyRadPlan.io import load_patient, load_tg119
from pyRadPlan.ct import CT
from pyRadPlan.cst import StructureSet


# Paths
@pytest.fixture
def tg119_path():
    return resources.files("pyRadPlan.data.phantoms").joinpath("TG119.mat")


def test_file_handler(tmp_path, tg119_path):
    """Test loading method."""

    tg119_load = matfile.load(tg119_path)

    tg119 = {
        "ct": tg119_load["ct"],
        "cst": tg119_load["cst"],
    }

    file_handler = MatlabFileHandler(tmp_path)
    file_handler.save(**tg119)

    assert os.path.exists(tmp_path.joinpath("ct.mat"))
    assert os.path.exists(tmp_path.joinpath("cst.mat"))

    ct = file_handler.load("ct")
    assert isinstance(ct, dict)
    cst = file_handler.load("cst")
    assert isinstance(cst, dict)

    file_handler.delete("ct", "cst")
    assert not os.path.exists(tmp_path.joinpath("ct.mat"))
    assert not os.path.exists(tmp_path.joinpath("cst.mat"))


def test_load_patient_from_file(tg119_path):
    ct, cst = load_patient(tg119_path)

    assert isinstance(ct, CT)
    assert isinstance(cst, StructureSet)


def test_load_tg119(tg119_path):
    ct, cst = load_patient(tg119_path)
    ct2, cst2 = load_tg119()

    assert isinstance(ct2, CT)
    assert isinstance(cst2, StructureSet)
    assert ct == ct2
    assert cst == cst2


def test_load_file_extradata(tg119_path):
    extra_data = {}
    extra_plan_data = {}
    ct, cst = load_patient(tg119_path, extra_data=extra_data, extra_plan_data=extra_plan_data)

    assert isinstance(ct, CT)
    assert isinstance(cst, StructureSet)
    assert isinstance(extra_data, dict)
    assert isinstance(extra_plan_data, dict)


def test_load_invalid_file():
    with pytest.raises(FileNotFoundError):
        _, _ = load_patient("unsupported_file.txt")



import pytest

from pyRadPlan.plan import PhotonPlan
from pyRadPlan.optimization.problems import (
    get_available_problems,
    get_problem,
    get_problem_from_pln,
    PlanningProblem,
)


def test_get_available_problems():
    problems = get_available_problems()
    assert isinstance(problems, dict)
    assert len(problems) > 0
    assert all([isinstance(k, str) for k in problems.keys()])
    assert all([issubclass(v, PlanningProblem) for v in problems.values()])


def test_get_available_problems_with_pln_default():
    pln = PhotonPlan()
    problems = get_available_problems(pln)
    assert isinstance(problems, dict)
    assert len(problems) > 0
    assert all([isinstance(k, str) for k in problems.keys()])


def test_get_problem():
    problems = get_available_problems()
    for problem_name, problem_class in problems.items():
        problem = get_problem(problem_name)
        assert isinstance(problem, problem_class)
        assert isinstance(problem, PlanningProblem)
        assert problem.short_name == problem_name
        assert problem.name == problem_class.name
        assert problem.short_name == problem_class.short_name


def test_get_problem_from_pln_specific():
    pln = PhotonPlan(prop_opt={"problem": "nonlin_fluence", "bypass_objective_jacobian": False})
    problem = get_problem_from_pln(pln)
    assert isinstance(problem, PlanningProblem)
    assert problem.short_name == "nonlin_fluence"
    assert not problem.bypass_objective_jacobian


def test_get_problem_from_pln_invalid_warning():
    pln = PhotonPlan(prop_opt={"problem": "invalid_problem"})
    with pytest.warns(UserWarning):
        get_problem_from_pln(pln)


def test_get_problem_invalid():
    with pytest.raises(KeyError):
        get_problem("invalid_problem")
    with pytest.raises(ValueError):
        get_problem(123)
    with pytest.raises(NotImplementedError):
        get_problem({"problem": "scipy"})  # valid problem, but dict not implemented yet



import pytest

from pyRadPlan.plan import validate_pln
from pyRadPlan.stf import validate_stf
from pyRadPlan.dij import validate_dij
from pyRadPlan.ct import validate_ct
from pyRadPlan.cst import validate_cst


@pytest.fixture
def test_data_photons(test_data_photons_raw):
    tmp = test_data_photons_raw

    pln = validate_pln(tmp["pln"])
    pln.prop_dose_calc["dosimetric_lateral_cutoff"] = 0.995
    pln.prop_dose_calc["lateral_model"] = "single"

    ct = validate_ct(tmp["ct"])
    cst = validate_cst(tmp["cst"], ct=ct)
    dij_matRad = validate_dij(tmp["dij"])
    stf = validate_stf(tmp["stf"])

    result = tmp["resultGUI"]

    return pln, ct, cst, stf, dij_matRad, result


@pytest.fixture
def test_data_protons(test_data_protons_raw):
    tmp = test_data_protons_raw

    pln = validate_pln(tmp["pln"])
    pln.prop_dose_calc["dosimetric_lateral_cutoff"] = 0.995
    pln.prop_dose_calc["lateral_model"] = "single"

    ct = validate_ct(tmp["ct"])
    cst = validate_cst(tmp["cst"], ct=ct)
    dij_matRad = validate_dij(tmp["dij"])
    stf = validate_stf(tmp["stf"])

    result = tmp["resultGUI"]

    return pln, ct, cst, stf, dij_matRad, result


@pytest.fixture
def test_data_helium(test_data_helium_raw):
    tmp = test_data_helium_raw

    pln = validate_pln(tmp["pln"])
    pln.prop_dose_calc["dosimetric_lateral_cutoff"] = 0.995
    pln.prop_dose_calc["lateral_model"] = "single"

    ct = validate_ct(tmp["ct"])
    cst = validate_cst(tmp["cst"], ct=ct)
    dij_matRad = validate_dij(tmp["dij"])
    stf = validate_stf(tmp["stf"])

    result = tmp["resultGUI"]

    return pln, ct, cst, stf, dij_matRad, result


@pytest.fixture
def test_data_carbon(test_data_carbon_raw):
    tmp = test_data_carbon_raw
    pln = validate_pln(tmp["pln"])
    pln.prop_dose_calc["dosimetric_lateral_cutoff"] = 0.995
    pln.prop_dose_calc["lateral_model"] = "single"

    ct = validate_ct(tmp["ct"])
    cst = validate_cst(tmp["cst"], ct=ct)
    dij_matRad = validate_dij(tmp["dij"])
    stf = validate_stf(tmp["stf"])

    result = tmp["resultGUI"]

    return pln, ct, cst, stf, dij_matRad, result



from typing import Any, Union

from pyRadPlan.machines import Machine, PhotonLINAC, IonAccelerator


def validate_machine(data: Union[dict[str, Any], Machine, None] = None, **kwargs) -> Machine:
    """
    Factory function to create a Machine object.

    Parameters
    ----------
    data : Union[Dict[str, Any], None]
        Dictionary containing the data to create the Machine object.
    **kwargs
        Arbitrary keyword arguments.

    Returns
    -------
    Machine
        A Machine object.

    Raises
    ------
    ValueError
        If the machine can not be constructed.
    """

    if data:
        # If data is already a Machine object, return it directly
        if isinstance(data, Machine):
            return data

        # Now we check for a dictionary
        if isinstance(data, dict):
            metadata = data.get("meta")

            if isinstance(metadata, dict):
                # obtain the radiation mode if we have a dictionary at our hands
                radiation_mode = metadata.get("radiation_mode")

                # Since we also allow camelCase, try to get radiationMode
                # if radiation_mode is not set
                if radiation_mode is None:
                    radiation_mode = metadata.get("radiationMode")

                if radiation_mode == "photons":
                    return PhotonLINAC.model_validate(data)

                if radiation_mode in ["protons", "helium", "carbon"]:
                    return IonAccelerator.model_validate(data)

        raise ValueError("Dictionary Structure of provided machine not valid!")

    radiation_mode = kwargs.get("radiation_mode", "")
    if radiation_mode == "photons":
        return PhotonLINAC(**kwargs)

    if radiation_mode in ["protons", "helium", "carbon"]:
        return IonAccelerator(**kwargs)

    raise ValueError(f"Unknown radiation mode: {radiation_mode}")



import logging

import numpy as np

from ._externalbeam import StfGeneratorExternalBeamRayBixel
from pyRadPlan.machines import PhotonLINAC

logger = logging.getLogger(__name__)


class StfGeneratorPhotonIMRT(StfGeneratorExternalBeamRayBixel):
    """Class representing a Photon IMRT Geometry Stf Generator.

    Attributes
    ----------
    name : str
        The name of the generator ("Photon IMRT Geometry").
    short_name : str
        The short name of the generator ("photonIMRT").
    possible_radiation_modes : list[str]
        A list of possible radiation modes (["photons"]).
    """

    name = "Photon IMRT Geometry"
    short_name = "photonIMRT"
    possible_radiation_modes = ["photons"]

    energy: float = 6.0

    def __init__(self, pln=None):
        self.radiation_mode = "photons"  # Radiation mode must be photons
        super().__init__(pln)

    def _initialize(self):
        super()._initialize()

        if not isinstance(self.machine, PhotonLINAC):
            raise ValueError("Machine must be an instance of PhotonLINAC.")

        energy_ix = self.machine.get_closest_energy_index(self.energy)

        machine_energy = self.machine.energies[energy_ix]

        if machine_energy != self.energy:
            logger.warning(
                "Selected energy not available in machine. Using closest available energy %g.",
                machine_energy,
            )
            self.energy = machine_energy

    def _generate_source_geometry(self):
        """Generate the source geometry for the photon IMRT geometry."""
        stf = super()._generate_source_geometry()
        return stf

    def _create_rays(self, beam: dict) -> list[dict]:
        """Create the rays for the photon IMRT geometry.

        Parameters
        ----------
        beam : dict
            The beam dictionary.

        Returns
        -------
        list[dict]
            A list of ray dictionaries
        """
        rays = super()._create_rays(beam)

        beamlet_template = {"energy": self.energy}
        rays = [
            {
                **ray,
                "beamlets": [beamlet_template.copy()],
                "target_point": 2 * ray["ray_pos"] - beam["source_point"],
                "target_point_bev": 2 * ray["ray_pos_bev"] - beam["source_point_bev"],
            }
            for ray in rays
        ]

        return rays


class StfGeneratorPhotonCollimatedSquareFields(StfGeneratorExternalBeamRayBixel):
    """Class representing a Photon Collimated Square Fields Stf Generator.

    Attributes
    ----------
    name : str
        The name of the generator ("Photon Collimated Square Fields").
    short_name : str
        The short name of the generator ("photonCollimatedSquareFields").
    possible_radiation_modes : list[str]
        A list of possible radiation modes (["photons"]).
    """

    name = "Photon Open Fields Geometry"
    short_name = "photonOpenFields"
    possible_radiation_modes = ["photons"]

    # Alias for field_width
    @property
    def field_width(self) -> float:
        """Alias for the bixel_width property."""
        return self.bixel_width

    @field_width.setter
    def field_width(self, value: float):
        self.bixel_width = value

    def __init__(self, pln=None):
        self.radiation_mode = "photons"
        super().__init__(pln)

    def _generate_ray_positions_in_isocenter_plane(self, beam):
        """Generate the ray positions in the isocenter plane.

        As we have a square field, this is a single ray to the isocenter.

        Parameters
        ----------
        beam : dict
            The beam dictionary. Ignored in this case

        Returns
        -------
        np.ndarray
            A numpy array with the single ray position of (0, 0, 0) in the isocenter.
        """

        return np.zeros((3, 1), dtype=float)

    def _generate_source_geometry(self):
        """Generate the source geometry for photon collimated square fields."""
        stf = super()._generate_source_geometry()

        for i, field in enumerate(stf):
            field["field_width"] = self.field_width

        return stf



import numpy as np
from pyRadPlan.geometry.lps import (
    get_beam_rotation_matrix,
    get_gantry_rotation_matrix,
    get_couch_rotation_matrix,
)


def test_get_gantry_rotation_matrix_90():
    gantry_angle = 90
    expected_matrix = np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]])
    result = get_gantry_rotation_matrix(gantry_angle)
    np.testing.assert_almost_equal(result, expected_matrix, decimal=6)


def test_get_gantry_rotation_matrix_0():
    gantry_angle = 0
    expected_matrix = np.eye(3)
    result = get_gantry_rotation_matrix(gantry_angle)
    np.testing.assert_almost_equal(result, expected_matrix, decimal=6)


def test_get_gantry_rotation_matrix_180():
    gantry_angle = 180
    expected_matrix = np.array([[-1, 0, 0], [0, -1, 0], [0, 0, 1]])
    result = get_gantry_rotation_matrix(gantry_angle)
    np.testing.assert_almost_equal(result, expected_matrix, decimal=6)


def test_get_gantry_rotation_matrix_360():
    gantry_angle = 360
    expected_matrix = np.eye(3)
    result = get_gantry_rotation_matrix(gantry_angle)
    np.testing.assert_almost_equal(result, expected_matrix, decimal=6)


def test_get_couch_rotation_matrix_90():
    couch_angle = 90
    expected_matrix = np.array([[0, 0, 1], [0, 1, 0], [-1, 0, 0]])
    result = get_couch_rotation_matrix(couch_angle)
    np.testing.assert_almost_equal(result, expected_matrix, decimal=6)


def test_get_couch_rotation_matrix_0():
    couch_angle = 0
    expected_matrix = np.eye(3)
    result = get_couch_rotation_matrix(couch_angle)
    np.testing.assert_almost_equal(result, expected_matrix, decimal=6)


def test_get_couch_rotation_matrix_180():
    couch_angle = 180
    expected_matrix = np.array([[-1, 0, 0], [0, 1, 0], [0, 0, -1]])
    result = get_couch_rotation_matrix(couch_angle)
    np.testing.assert_almost_equal(result, expected_matrix, decimal=6)


def test_get_couch_rotation_matrix_360():
    couch_angle = 360
    expected_matrix = np.eye(3)
    result = get_couch_rotation_matrix(couch_angle)
    np.testing.assert_almost_equal(result, expected_matrix, decimal=6)


def test_get_beam_rotation_matrix_90_45():
    gantry_angle = 90
    couch_angle = 45
    a45 = np.sin(np.deg2rad(45))
    expected_matrix = np.array([[0.0, -a45, a45], [1.0, 0.0, 0.0], [0.0, a45, a45]])
    result = get_beam_rotation_matrix(gantry_angle, couch_angle)
    np.testing.assert_almost_equal(result, expected_matrix, decimal=6)


def test_get_beam_rotation_matrix_0_0():
    gantry_angle = 0
    couch_angle = 0
    expected_matrix = np.eye(3)
    result = get_beam_rotation_matrix(gantry_angle, couch_angle)
    np.testing.assert_almost_equal(result, expected_matrix, decimal=6)


def test_get_beam_rotation_matrix_45_90():
    gantry_angle = 45
    couch_angle = 90
    a45 = np.sin(np.deg2rad(45))
    expected_matrix = np.array([[0.0, 0, 1.0], [a45, a45, 0.0], [-a45, a45, 0.0]])
    result = get_beam_rotation_matrix(gantry_angle, couch_angle)
    np.testing.assert_almost_equal(result, expected_matrix, decimal=6)


def test_get_beam_rotation_matrix_360_360():
    gantry_angle = 360
    couch_angle = 360
    expected_matrix = np.eye(3)
    result = get_beam_rotation_matrix(gantry_angle, couch_angle)
    np.testing.assert_almost_equal(result, expected_matrix, decimal=6)



import pytest
import sys

if sys.version_info < (3, 10):
    import importlib_resources as resources  # Backport for older versions
else:
    from importlib import resources  # Standard from Python 3.9+
import pymatreader
import numpy as np
import SimpleITK as sitk
from pyRadPlan.ct import create_ct


@pytest.fixture
def sample_image_3d():
    """Create a sample SimpleITK image for testing."""
    dims = [50, 100, 100]
    sample_array = np.ones(dims) * 1000  # Random HU values
    image = sitk.GetImageFromArray(sample_array)
    image.SetOrigin((0, 0, 0))
    image.SetSpacing((1, 1, 2))
    image.SetDirection((1, 0, 0, 0, 1, 0, 0, 0, 1))
    return image, dims


@pytest.fixture
def sample_image_4d(sample_image_3d):
    image, dims_3d = sample_image_3d
    image_4d = sitk.JoinSeries([image, image])
    image_4d.SetOrigin((0, 0, 0, 0))
    image_4d.SetSpacing((1, 1, 2, 1))
    image_4d.SetDirection((1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1))
    return image_4d, dims_3d


@pytest.fixture
def generic_input_3d(sample_image_3d):
    cube_hu, dims = sample_image_3d
    mask = np.zeros(dims).astype("uint8")
    mask[0, 0, 1] = 1
    return "VOI", create_ct(cube_hu=cube_hu), mask, 0.2, 0.3


@pytest.fixture
def generic_input_4d(sample_image_4d):
    cube_hu, dims_3d = sample_image_4d
    mask_1 = np.zeros(dims_3d, dtype=np.uint8)
    mask_1[0, 0, 1] = 1
    mask_2 = np.zeros(dims_3d, dtype=np.uint8)
    mask_2[1, 1, 0] = 1
    mask_np = [sitk.GetImageFromArray(mask_1, False), sitk.GetImageFromArray(mask_2, False)]
    mask = sitk.JoinSeries(mask_np)
    return "VOI", create_ct(cube_hu=cube_hu), mask, 0.2, 0.3


@pytest.fixture
def matrad_import():
    """Load a matRad phantom for testing."""
    files = resources.files("pyRadPlan.data.phantoms")
    path = files.joinpath("TG119.mat")
    tmp = pymatreader.read_mat(path)
    return tmp



"""Image / Grid Resampling."""

import numpy as np
import SimpleITK as sitk
from ._grids import Grid


def resample_image(
    input_image: sitk.Image,
    interpolator=sitk.sitkBSpline,
    target_image: sitk.Image = None,
    target_grid_spacing: tuple[float, float, float] = None,
    target_grid: Grid = None,
) -> sitk.Image:  # also accept sitk images and grid points. Can be dealt with with kwargs
    """
    Resample an sitk Image.

    Use target resolution, reference image, dimensions, or Grid as input.

    Parameters
    ----------
    input_image : sitk.Image
        The input image to be resampled.
    interpolator : sitk.InterpolatorEnum, optional
        The interpolator to use for resampling. Default is sitk.sitkBSpline.
    target_image : sitk.Image, optional
        The reference image to resample to. Default is None.
    target_grid_spacing : tuple[float,float,float], optional
        The target grid spacing to resample to. Default is None.
    target_grid : Grid, optional
        The target grid to resample to. Default is None.

    Returns
    -------
    sitk.Image
        The resampled image.

    Notes
    -----
    Exactly one of target_image, target_grid_spacing, or target_grid must be provided.
    """

    if [target_image, target_grid_spacing, target_grid].count(None) != 2:
        raise TypeError(
            "Only one of target_image, target_grid_spacing, or target_grid must be provided."
        )

    image_grid = Grid.from_sitk_image(input_image)

    if target_image is not None:
        target_grid = Grid.from_sitk_image(target_image)
    elif target_grid_spacing is not None:
        target_grid = image_grid.resample(target_grid_spacing)

    resample = sitk.ResampleImageFilter()
    resample.SetOutputSpacing(target_grid.resolution_vector)
    resample.SetSize(target_grid.dimensions)
    resample.SetOutputDirection(target_grid.direction.ravel())
    resample.SetOutputOrigin(target_grid.origin)
    resample.SetInterpolator(interpolator)
    resampled_image = resample.Execute(input_image)

    return resampled_image


def resample_numpy_array(
    input_array: np.ndarray,
    reference_image: sitk.Image = None,
    reference_grid: Grid = None,
    interpolator=sitk.sitkBSpline,
    target_image: sitk.Image = None,
    target_grid_spacing: tuple[float, float, float] = None,
    target_grid: Grid = None,
) -> np.ndarray:
    """
    Resample a numpy grid.

    Use target resolution, reference image, dimensions, or Grid as input.

    We convert the numpy array to an sitk.Image, so the dimensions will be
    switched. More exactly,
    the numpy array will be index i,j,k <-> z,y,x, which will converted to
    j,i,k <-> y,x,z in sitk.
    This is to be considered when supplying the reference grid  or image.

    Parameters
    ----------
    input_image : sitk.Image
        The input image to be resampled.
    reference_image : sitk.Image, optional
        The reference image providing spatial information of the array. Default is None.
        Be wary of sitk <-> numpy indexing conventions.
    reference_grid : Grid, optional
        The reference grid of the resampled array. Default is None.
        Be wary of sitk <-> numpy indexing conventions.
    interpolator : sitk.InterpolatorEnum, optional
        The interpolator to use for resampling. Default is sitk.sitkBSpline.
    target_image : sitk.Image, optional
        The reference image to resample to. Default is None.
    target_grid_spacing : tuple[float,float,float], optional
        The target grid spacing to resample to. Default is None.
    target_grid : Grid, optional
        The target grid to resample to. Default is None.

    Returns
    -------
    sitk.Image
        The resampled image.

    Notes
    -----
    Exactly one of target_image, target_grid_spacing, or target_grid must be provided.
    """

    if [reference_image, reference_grid].count(None) != 1:
        raise TypeError("Only one of reference_image or reference_grid must be provided.")

    if reference_image is not None:
        reference_grid = Grid.from_sitk_image(reference_image)

    # I think there is an issue here. The axes are flipped with GetImageFromArray
    array_sitk = sitk.GetImageFromArray(input_array)
    array_sitk.SetOrigin(reference_grid.origin)
    array_sitk.SetSpacing(reference_grid.resolution_vector)
    array_sitk.SetDirection(reference_grid.direction.ravel())

    resampled_array_sitk = resample_image(
        input_image=array_sitk,
        interpolator=interpolator,
        target_image=target_image,
        target_grid_spacing=target_grid_spacing,
        target_grid=target_grid,
    )

    return sitk.GetArrayFromImage(resampled_array_sitk)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import os
import sys
from pyRadPlan import __version__

print(os.path.abspath("../pyRadPlan"))
sys.path.insert(0, os.path.abspath("../pyRadPlan"))  # Adjust to your source folder

project = "pyRadPlan"
copyright = "2024, e0404"
author = "e0404"

version = __version__
release = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "numpydoc",
    "sphinx_autodoc_typehints",
    "sphinxcontrib.autodoc_pydantic",
    "sphinx_design",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

add_module_names = False


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]

autodoc_type_aliases = {
    "npt.ArrayLike": "npt.ArrayLike",
}

autodoc_default_options = {
    "no-module": True,  # Suppress the module/package labels
}

numpydoc_class_members_toctree = False

# disable show json as it otherwises crashes the documentation building at the moment
autodoc_pydantic_model_show_json = False



import pytest
import sys

if sys.version_info < (3, 10):
    import importlib_resources as resources  # Backport for older versions
else:
    from importlib import resources  # Standard from Python 3.9+

import numpy as np
import SimpleITK as sitk

from pyRadPlan.io import load_patient
from pyRadPlan.plan import create_pln
from pyRadPlan.ct import CT, validate_ct
from pyRadPlan.cst import validate_cst, StructureSet, VOI

from pyRadPlan.stf import Ray, get_available_generators, get_generator
from pyRadPlan.stf.generators import (
    StfGeneratorBase,
    StfGeneratorExternalBeamRayBixel,
    StfGeneratorPhotonIMRT,
    StfGeneratorPhotonCollimatedSquareFields,
    StfGeneratorIonSingleSpot,
    StfGeneratorIMPT,
)


@pytest.fixture
def sample_ct():
    image = sitk.GetImageFromArray(np.random.rand(5, 15, 25) * 1000)  # Random HU values
    image.SetOrigin((0, 0, 0))
    image.SetSpacing((2, 3, 4))  # Irregular spacing for test
    image.SetDirection((1, 0, 0, 0, 1, 0, 0, 0, 1))

    ct = CT(cube_hu=image)

    return ct


@pytest.fixture
def sample_cst(sample_ct):
    organ_mask = np.zeros((5, 15, 25), dtype=np.uint8)
    organ_mask[1:3, 3:11, 5:20] = 1
    organ_mask = sitk.GetImageFromArray(organ_mask)

    target_mask = np.zeros((5, 15, 25), dtype=np.uint8)
    target_mask[2, 5:9, 10:15] = 1
    target_mask = sitk.GetImageFromArray(target_mask)

    target1 = VOI(name="testtarget", voi_type="TARGET", mask=target_mask, ct_image=sample_ct)
    organ1 = VOI(name="testorgan", voi_type="OAR", mask=organ_mask, ct_image=sample_ct)

    cst = StructureSet(vois=[target1, organ1], ct_image=sample_ct)

    return cst


@pytest.fixture
def tg119():
    path = resources.files("pyRadPlan.data.phantoms")
    ct, cst = load_patient(path.joinpath("TG119.mat"))
    return {"ct": ct, "cst": cst}


@pytest.fixture
def sample_photon_pln_dict():
    pln = {
        "radiation_mode": "photons",
        "prop_stf": {"bixel_width": 2.0},
        "machine": "Generic",
    }
    return pln


@pytest.fixture
def sample_photon_pln(sample_photon_pln_dict):
    pln = create_pln(sample_photon_pln_dict)
    return pln


@pytest.fixture
def sample_proton_pln_dict():
    pln = {
        "radiation_mode": "protons",
        "prop_stf": {"bixel_width": 2.0},
        "machine": "Generic",
    }
    return pln


@pytest.fixture
def sample_proton_pln(sample_proton_pln_dict):
    pln = create_pln(sample_proton_pln_dict)
    return pln


def test_available_generators(sample_proton_pln):
    generators = get_available_generators(sample_proton_pln)
    assert isinstance(generators, dict)
    assert len(generators.items()) > 0


def test_get_generator(sample_proton_pln):
    sample_proton_pln.prop_stf["generator"] = "IMPT"
    generator = get_generator(sample_proton_pln)
    assert generator
    assert isinstance(generator, StfGeneratorBase)


def test_get_generator_default(sample_proton_pln):
    generator = get_generator(sample_proton_pln)
    assert generator
    assert isinstance(generator, StfGeneratorBase)


def test_get_generator_invalid(sample_proton_pln):
    sample_proton_pln.prop_stf["generator"] = "InvalidGenerator"
    with pytest.warns(UserWarning):
        generator = get_generator(sample_proton_pln)


def test_basic_photon_imrt_construct(sample_photon_pln_dict, sample_photon_pln):
    stf_gen = StfGeneratorPhotonIMRT()
    assert isinstance(
        stf_gen, (StfGeneratorBase, StfGeneratorExternalBeamRayBixel, StfGeneratorPhotonIMRT)
    )

    stf_gen = StfGeneratorPhotonIMRT(sample_photon_pln)
    assert isinstance(
        stf_gen, (StfGeneratorBase, StfGeneratorExternalBeamRayBixel, StfGeneratorPhotonIMRT)
    )
    assert stf_gen.bixel_width == sample_photon_pln.prop_stf["bixel_width"]

    stf_gen = StfGeneratorPhotonIMRT(sample_photon_pln_dict)
    assert isinstance(
        stf_gen, (StfGeneratorBase, StfGeneratorExternalBeamRayBixel, StfGeneratorPhotonIMRT)
    )
    assert stf_gen.bixel_width == sample_photon_pln.prop_stf["bixel_width"]


def test_basic_photon_imrt(sample_ct, sample_cst, sample_photon_pln):
    stf_gen = StfGeneratorPhotonIMRT(sample_photon_pln)

    stf_gen.gantry_angles = [90.0, 270.0]
    stf_gen.couch_angles = [0.0, 90.0]

    stf = stf_gen.generate(sample_ct, sample_cst)

    assert len(stf) == 2
    for i in range(len(stf)):
        assert isinstance(stf[i], dict)
        assert stf[i]["gantry_angle"] == stf_gen.gantry_angles[i]
        assert stf[i]["couch_angle"] == stf_gen.couch_angles[i]
        assert stf[i]["bixel_width"] == stf_gen.bixel_width
        assert "rays" in stf[i]
        assert isinstance(stf[i]["rays"], list)

        for ray in stf[i]["rays"]:
            assert isinstance(ray, dict)
            assert "ray_pos_bev" in ray
            assert isinstance(ray["ray_pos_bev"], np.ndarray)
            assert ray["ray_pos_bev"].size == 3


def test_basic_photon_collimated_field(sample_ct, sample_cst, sample_photon_pln):
    stf_gen = StfGeneratorPhotonCollimatedSquareFields(sample_photon_pln)

    stf_gen.gantry_angles = [0.0, 270.0]
    stf_gen.couch_angles = [0.0, 90.0]

    expected_source_points_bev = [
        np.array([0, -1000, 0], dtype=float),
        np.array([0, -1000, 0], dtype=float),
    ]

    expected_source_points = [
        np.array([0, -1000.0, 0], dtype=float),
        np.array(
            [
                0,
                0,
                1000.0,
            ],
            dtype=float,
        ),
    ]

    expected_ray_pos_bev = [np.array([0, 0, 0], dtype=float), np.array([0, 0, 0], dtype=float)]

    expected_ray_pos = [
        np.array([[0, 0, 0], [0, 0, 0]], dtype=float),
        np.array([[0, 0, 0], [0, 0, 0]], dtype=float),
    ]

    stf = stf_gen.generate(sample_ct, sample_cst)

    assert len(stf) == 2
    for i, field in enumerate(stf):
        assert isinstance(field, dict)
        assert field["gantry_angle"] == stf_gen.gantry_angles[i]
        assert field["couch_angle"] == stf_gen.couch_angles[i]
        assert "field_width" in field
        assert "bixel_width" in field
        assert field["bixel_width"] == stf_gen.bixel_width
        assert "source_point" in field
        assert isinstance(field["source_point"], np.ndarray)
        assert field["source_point"].size == 3
        assert "source_point_bev" in field
        assert isinstance(field["source_point_bev"], np.ndarray)
        assert field["source_point_bev"].size == 3
        assert np.isclose(field["source_point_bev"], expected_source_points_bev[i]).all()
        assert np.isclose(field["source_point"], expected_source_points[i]).all()
        assert np.isclose(field["rays"][0]["ray_pos_bev"], expected_ray_pos_bev[i]).all()
        assert np.isclose(field["rays"][0]["ray_pos"], expected_ray_pos[i]).all()


def test_single_ion_spot(sample_ct, sample_cst, sample_proton_pln):
    stf_gen = StfGeneratorIonSingleSpot(sample_proton_pln)

    stf_gen.gantry_angles = [0.0, 270.0]
    stf_gen.couch_angles = [0.0, 90.0]
    sad = 10000.0
    expected_source_points_bev = [
        np.array([0, -sad, 0], dtype=float),
        np.array([0, -sad, 0], dtype=float),
    ]

    expected_source_points = [
        np.array([0, -sad, 0], dtype=float),
        np.array(
            [
                0,
                0,
                sad,
            ],
            dtype=float,
        ),
    ]

    expected_ray_pos_bev = [np.array([0, 0, 0], dtype=float), np.array([0, 0, 0], dtype=float)]

    expected_ray_pos = [
        np.array([[0, 0, 0], [0, 0, 0]], dtype=float),
        np.array([[0, 0, 0], [0, 0, 0]], dtype=float),
    ]

    stf = stf_gen.generate(sample_ct, sample_cst)

    assert len(stf) == 2
    for i, field in enumerate(stf):
        assert isinstance(field, dict)
        assert field["gantry_angle"] == stf_gen.gantry_angles[i]
        assert field["couch_angle"] == stf_gen.couch_angles[i]
        assert "bixel_width" in field
        assert field["bixel_width"] == stf_gen.bixel_width
        assert "source_point" in field
        assert isinstance(field["source_point"], np.ndarray)
        assert field["source_point"].size == 3
        assert "source_point_bev" in field
        assert isinstance(field["source_point_bev"], np.ndarray)
        assert field["source_point_bev"].size == 3
        assert np.isclose(field["source_point_bev"], expected_source_points_bev[i]).all()
        assert np.isclose(field["source_point"], expected_source_points[i]).all()
        assert np.isclose(field["rays"][0]["ray_pos_bev"], expected_ray_pos_bev[i]).all()
        assert np.isclose(field["rays"][0]["ray_pos"], expected_ray_pos[i]).all()


def test_impt(sample_ct, sample_cst, sample_proton_pln):
    stf_gen = StfGeneratorIMPT(sample_proton_pln)

    stf_gen.gantry_angles = [0.0, 270.0]
    stf_gen.couch_angles = [0.0, 90.0]

    stf = stf_gen.generate(sample_ct, sample_cst)

    assert len(stf) == 2
    for i, field in enumerate(stf):
        assert isinstance(field, dict)
        assert field["gantry_angle"] == stf_gen.gantry_angles[i]
        assert field["couch_angle"] == stf_gen.couch_angles[i]
        assert "bixel_width" in field
        assert field["bixel_width"] == stf_gen.bixel_width
        assert "source_point" in field
        assert isinstance(field["source_point"], np.ndarray)
        assert field["source_point"].size == 3
        assert "source_point_bev" in field
        assert isinstance(field["source_point_bev"], np.ndarray)
        assert field["source_point_bev"].size == 3
        assert len(field["rays"]) > 0
        for ray in field["rays"]:
            assert isinstance(ray, Ray)


def test_impt_tg119(tg119, sample_proton_pln):
    ct = tg119["ct"]
    cst = tg119["cst"]

    stf_gen = StfGeneratorIMPT(sample_proton_pln)
    stf_gen.bixel_width = 5.0
    stf = stf_gen.generate(ct, cst)

    assert isinstance(stf, list)
    assert len(stf) == 1

    for i, field in enumerate(stf):
        assert isinstance(field, dict)
        assert "bixel_width" in field
        assert field["bixel_width"] == stf_gen.bixel_width
        assert "source_point" in field
        assert isinstance(field["source_point"], np.ndarray)
        assert field["source_point"].size == 3
        assert "source_point_bev" in field
        assert isinstance(field["source_point_bev"], np.ndarray)
        assert field["source_point_bev"].size == 3
        assert len(field["rays"]) > 0
        for ray in field["rays"]:
            assert isinstance(ray, Ray)



from typing import Any, Union
from typing_extensions import Self
import numpy as np
from pydantic import (
    SerializationInfo,
    model_validator,
    field_serializer,
    ValidatorFunctionWrapHandler,
    ValidationInfo,
    ValidationError,
)
from numpydantic import NDArray, Shape
from pyRadPlan.stf._beam import Beam
from pyRadPlan.core import PyRadPlanBaseModel
from pyRadPlan.util.helpers import models2recarray


class SteeringInformation(PyRadPlanBaseModel):
    """
    A class representing the Steering Information (stf).

    This class extends PyRadPlanBaseModel (based on pydantic) and provides functionality to
    handle single beams, including their properties.
    These are defined in the corresponding class (_Beam.py).

    Attributes
    ----------
    beams : List[Beam] - list consisting of Beam objects (pydantic)
        beam class object containing the properties of the beam.

    Methods
    -------
    validate_model_input(data: Any) -> Any
        Validates the input data before creating the model instance.

    to_matrad() -> dict
        Creates a dictionary ready to save the stf model to a mat-file that can be read.
    """

    beams: list[Beam]

    # Validation
    @model_validator(mode="wrap")
    @classmethod
    def validate_model_input(
        cls, data: Any, handler: ValidatorFunctionWrapHandler, info: ValidationInfo
    ) -> Self:
        """
        Validate the input data for creating the model instance.

        Will first try to run pydantics handler, and if it fails with a
        ValidationError, it will
        try to convert the data to the right format.
        """
        try:
            return handler(data, info)
        except ValidationError:
            # Check if import is from matlab
            # If from matlab but stf already choosen, the code is executed as normal
            matrad_format = ["__header__", "__version__", "__globals__"]
            if all(key in data for key in matrad_format):
                # struct of matlab usual in 4. position. #TODO: are there exceptions?
                data = data[list(data.keys())[3]]

            if isinstance(data, dict) and "beams" not in data:
                # This code is needed to pass the right format of the stf from matRad
                try:
                    length_lists = [len(v) for v in data.values()]
                except (TypeError, ValueError):
                    return handler({"beams": [data]}, info)  # TODO: only works if its a list??

                if len(set(length_lists)) == 1 and len(length_lists) > 1:
                    beam_list = []
                    length_lists = []

                    for v in data.values():
                        if isinstance(v, list):
                            length_lists.append(len(v))

                        # this exception is necessary if imported stf from matlab is only one beam
                        if isinstance(v, int or float):
                            length_lists.append(1)

                    for i in range(length_lists[0]):
                        beam = []

                        for key in data.keys():
                            entry = [key, data[key][i]]
                            beam.append(entry)

                        beam_list.append(dict(beam))

                    return handler({"beams": beam_list}, info)

            if isinstance(data, list):
                return handler({"beams": data}, info)

            return handler(data, info)
        except Exception as exc:
            raise exc

    @field_serializer("beams")
    def custom_beams_serializer(self, v: list[Beam], info: SerializationInfo) -> Any:
        """Serialize the beams fields in various contexts."""
        context = info.context
        if context and context.get("matRad") == "mat-file":
            override_types = {"rays": np.recarray}
            beams_recarray = models2recarray(
                v, override_types=override_types, serialization_context=context
            )
            return beams_recarray
        return [
            beam.model_dump(
                by_alias=info.by_alias,
            )
            for beam in v
        ]

    def to_matrad(self, context: str = "mat-file") -> Any:
        export = super().to_matrad(context=context)
        return export["beams"]

    @property
    def num_of_beams(self) -> int:
        return len(self.beams)

    @property
    def num_of_rays(self) -> int:
        return sum([beam.num_of_rays for beam in self.beams])

    @property
    def total_number_of_bixels(self) -> int:
        return sum([beam.total_number_of_bixels for beam in self.beams])

    @property
    def bixel_beam_index_map(self) -> NDArray[Shape["1-*"], np.int64]:
        """Mapping of bixels to their respective beam index."""
        tmp_map = np.zeros(self.total_number_of_bixels, dtype=np.int64)
        start = 0
        for b, beam in enumerate(self.beams):
            tmp_map[start : start + beam.total_number_of_bixels] = b
            start += beam.total_number_of_bixels
        return tmp_map

    @property
    def bixel_ray_index_per_beam_map(self) -> NDArray[Shape["1-*"], np.int64]:
        """Mapping of bixels to the ray index in the individual beams."""
        tmp_map = np.zeros(self.total_number_of_bixels, dtype=np.int64)
        start = 0
        for beam in self.beams:
            tmp_map[start : start + beam.total_number_of_bixels] = beam.bixel_ray_map
            start += beam.total_number_of_bixels
        return tmp_map

    @property
    def bixel_index_per_beam_map(self) -> NDArray[Shape["1-*"], np.int64]:
        """Mapping of bixels to their bixel index in the respective beam."""
        tmp_map = np.zeros(self.total_number_of_bixels, dtype=np.int64)
        start = 0
        for beam in self.beams:
            tmp_map[start : start + beam.total_number_of_bixels] = np.arange(
                beam.total_number_of_bixels
            )
            start += beam.total_number_of_bixels
        return tmp_map


def create_stf(
    stf: Union[dict[str, Any], SteeringInformation, None] = None, **kwargs
) -> SteeringInformation:
    """
    Create a Steering Information object.

    Parameters
    ----------
    stf : Union[dict[str, Any], None]
        dictionary containing the data to create the stf object.
    **kwargs
        Arbitrary keyword arguments.

    Returns
    -------
    SteeringInformation
        A SteeringInformation class object.

    Raises
    ------
    ValueError
        If the radiation mode is unknown or empty.
    """
    if stf:
        # If data is already a Stf object, return it directly
        if isinstance(stf, SteeringInformation):
            return stf
        return SteeringInformation.model_validate(stf)

    return SteeringInformation(**kwargs)  # not tested


def validate_stf(
    stf: Union[dict[str, Any], SteeringInformation, None] = None, **kwargs
) -> SteeringInformation:
    """
    Validate a Steering Information object.

    Synonym to create_stf but should be used in validation context.

    Parameters
    ----------
    stf : Union[dict[str, Any], None]
        dictionary containing the data to create the stf object.

    Returns
    -------
    SteeringInformation
        A validated SteeringInformation class object.
    """
    return create_stf(stf, **kwargs)



import pytest
import numpy as np

from pyRadPlan.optimization.solvers import get_solver, OptimizerIpopt, SolverBase

if OptimizerIpopt is None:
    pytest.skip("IPOPT not installed", allow_module_level=True)


def test_get_solver_ipopt():
    solver = get_solver("ipopt")
    assert isinstance(solver, OptimizerIpopt)
    assert isinstance(solver, SolverBase)
    assert solver.short_name == "ipopt"


def test_simple_problem_ipopt():
    solver = get_solver("ipopt")

    # Define the problem
    def objective(x):
        return float(x[0] ** 2 + x[1] ** 2)

    def gradient(x):
        return np.asarray([2 * x[0], 2 * x[1]], dtype=np.float64)

    solver.objective = objective
    solver.gradient = gradient

    # Initial guess
    x0 = np.asarray([1.0, 1.0], dtype=np.float64)

    # Solve
    result, status = solver.solve(x0)

    assert isinstance(result, np.ndarray)
    assert isinstance(status, int)

    assert np.all(np.isclose(result, 0.0, atol=1e-4))



from typing import Any
import numpy as np
from pydantic import (
    Field,
    model_validator,
    field_validator,
    computed_field,
    field_serializer,
    SerializationInfo,
)

from numpydantic import NDArray, Shape

from pyRadPlan.stf._ray import Ray
from pyRadPlan.core import PyRadPlanBaseModel
from pyRadPlan.util.helpers import models2recarray


class Beam(PyRadPlanBaseModel):
    """
    A class representing a single beam.

    This class extends PyRadPlanBaseModel (Pydantic) and provides functionality to
    handle the steering information, including properties like gantry angle, couch angle, etc.

    Attributes
    ----------
    gantry_angle :
        The gantry angle of the beam in (°).
    couch_angle :
        The couch angle of the beam in (°).
    bixel_width :
        The width of the bixels in (mm).
    radiation_mode :
        The radiation mode of the beam (e.g. photon, proton, carbon).
    machine :
        The machine used for the beam. (e.g. 'Generic')
    sad :
        The source to axis distance in (mm).
    iso_center :
        The isocenter of the beam in (x, y, z) coordinates.
    num_of_rays :
        The number of rays in the beam.
    ray :
        dict containing the properties of each ray.
    source_point_bev :
        The source point in BEV coordinates.
    source_point :
        The source point in (x, y, z) coordinates.
    num_of_bixels_per_ray :
        The number of bixels per ray in an array (1 x num_of_rays).
    longitudinal_spot_spacing :
        The longitudinal spot spacing in (mm).
    total_number_of_bixels :
        The total number of bixels in the beam.

    Methods
    -------
    create_beam():
        Validate and create a Beam object.
    validate_beam():
        Validate and create a Beam object.
    """

    gantry_angle: float = Field(default=0)
    couch_angle: float = Field(default=0)  # , gt=-100, lt=100)
    bixel_width: float = Field(default=5)
    radiation_mode: str = Field(default="protons")
    machine: str = Field(default="Generic")
    sad: float = Field(alias="SAD", default=100000)
    iso_center: NDArray[Shape["3"], np.float64]
    rays: list[Ray] = Field(
        alias="ray"
    )  # alias needed for matRad import. Can also be done in the model_validator

    source_point_bev: NDArray[Shape["3"], np.float64] = Field(
        alias="sourcePoint_bev", default=([0, -10000, 0]), validate_default=True
    )
    source_point: NDArray[Shape["3"], np.float64] = Field(
        default=([0, 0, 0]), validate_default=True
    )
    longitudinal_spot_spacing: float = Field(default=2.0)

    @model_validator(mode="before")
    @classmethod
    def validate_model_input(cls, data: Any) -> Any:
        # isinstance needed to exclude integers
        if isinstance(data, dict):
            if "ray" in data and isinstance(data["ray"], dict):
                ray_dict = data["ray"]

                # these are computed properties in python. Can be removed
                if "totalNumOfBixels" in ray_dict:
                    ray_dict.pop("totalNumOfBixels")
                if "numOfBixelsPerRay" in ray_dict:
                    ray_dict.pop("numOfBixelsPerRay")
                if "numOfRays" in ray_dict:
                    ray_dict.pop("numOfRays")

                length_lists = []

                for v in ray_dict.values():
                    if isinstance(v, list):
                        length_lists.append(len(v))

                    # this exception is necessary if imported stf from matlab is only one beam
                    # TODO: Test if this is necessary
                    if isinstance(v, int or float):
                        length_lists.append(1)

                # TODO: This error message might be confusing in some cases
                if len(set(length_lists)) != 1:
                    raise ValueError(
                        "All values in the 'ray' dictionary must have the same length"
                    )

                if len(set(length_lists)) == 1:
                    ray_list = []

                    for i in range(length_lists[0]):
                        ray = []

                        for key in ray_dict.keys():
                            entry = [key, ray_dict[key][i]]
                            ray.append(entry)

                        ray_list.append(dict(ray))

                    data.pop("ray")
                    data["rays"] = ray_list
        return data

    @field_validator("source_point", "source_point_bev", "iso_center", mode="before")
    @classmethod
    def validate_nparray_dtype(cls, v: Any) -> Any:
        """Validate arrays to have floating point values."""
        v = np.asarray(v, dtype=np.float64)
        return v.reshape((3,))

    @field_validator("rays", mode="after")
    @classmethod
    def validate_rays(cls, v: list[Ray]) -> list[Ray]:
        """
        Validate the rays attribute.

        Pydantic will validate that it is a list but will also allow list of dicts
        """
        try:
            return [Ray.model_validate(ray) for ray in v]
        except TypeError as e:
            raise ValueError(f"Error validating rays: {e}")

    @computed_field
    @property
    def num_of_bixels_per_ray(self) -> np.ndarray:
        return np.array([len(ray.beamlets) for ray in self.rays])

    @computed_field
    @property
    def num_of_rays(self) -> int:
        return len(self.rays)

    @computed_field(alias="totalNumOfBixels")
    @property
    def total_number_of_bixels(self) -> int:
        return int(sum(self.num_of_bixels_per_ray))

    @property
    def bixel_ray_map(self) -> NDArray[Shape["1-*"], np.int64]:
        """Map providing ray index in the beam for each bixel."""
        return np.repeat(np.arange(len(self.rays)), self.num_of_bixels_per_ray)

    # serialization
    @field_serializer("rays")
    def custom_rays_serializer(self, v: list[Ray], info: SerializationInfo) -> Any:
        context = info.context
        if context and context.get("matRad") == "mat-file":
            helper_model = Ray.create_matrad_helper_model()
            override_types = {"range_shifter": np.recarray, "beamlets": None}
            rays_matrad = [helper_model.model_validate(ray) for ray in v]
            rays_recarray = models2recarray(
                rays_matrad, override_types=override_types, serialization_context=context
            )
            # override_types = get_type_hints(Beamlet)
            # rays_recarray = models2recarray(v, serialization_context=context)
            return rays_recarray
        return [ray.model_dump(by_alias=info.by_alias) for ray in v]


def create_beam():
    """Validate and create a Beam object."""
    # TODO
    pass


def validate_beam():
    """Validate and create a Beam object."""
    # TODO
    pass



import numpy as np
from pyRadPlan.optimization.objectives import (
    DoseUniformity,
    SquaredDeviation,
    SquaredOverdosing,
    SquaredUnderdosing,
    EUD,
    MeanDose,
    MinDVH,
    MaxDVH,
    get_available_objectives,
    get_objective,
)


def test_objective_availability():
    available_objectives = get_available_objectives()
    assert "Dose Uniformity" in available_objectives
    assert "Squared Deviation" in available_objectives
    assert "Squared Overdosing" in available_objectives
    assert "Squared Underdosing" in available_objectives
    assert "EUD" in available_objectives
    assert "Mean Dose" in available_objectives
    assert "Min DVH" in available_objectives
    assert "Max DVH" in available_objectives


def test_get_objective_str():
    dose_uni = get_objective("Dose Uniformity")
    assert isinstance(dose_uni, DoseUniformity)


def test_get_objective_dict():
    dose_uni = get_objective({"name": "Dose Uniformity", "priority": 10.0})
    assert isinstance(dose_uni, DoseUniformity)
    assert dose_uni.priority == 10.0


def test_get_objective_instance():
    dose_uni = DoseUniformity(priority=10.0)
    dose_uni2 = get_objective(dose_uni)
    assert dose_uni == dose_uni2


def test_get_objective_from_matrad_tg119(tg119_raw):
    _, cst = tg119_raw

    obj_mat_1 = cst[0][5]
    obj = get_objective(obj_mat_1)
    obj_mat_2 = cst[1][5]
    obj2 = get_objective(obj_mat_2)

    assert isinstance(obj, SquaredOverdosing)
    assert obj.priority == obj_mat_1["penalty"]
    assert obj.d_max == obj_mat_1["parameters"]

    assert isinstance(obj2, SquaredDeviation)
    assert obj2.priority == obj_mat_2["penalty"]
    assert obj2.d_ref == obj_mat_2["parameters"]


def test_DoseUniformity_constructor():
    doseUni = DoseUniformity()
    assert doseUni.name == "Dose Uniformity"
    assert doseUni.parameters == []
    assert doseUni.priority == 1.0


def test_DoseUniformity_compute_objective():
    doseUni = DoseUniformity()

    dose = np.array([1, 2, 3])
    assert np.abs(doseUni.compute_objective(dose) - 1) < 1e-10


def test_DoseUniformity_compute_gradient():
    doseUni = DoseUniformity()
    dose = np.array([1, 2, 3])
    grad_expected = 1 / 2 * np.array([-1, 0, 1])
    assert np.all((doseUni.compute_gradient(dose) - grad_expected) < 1e-10)


def test_SquaredDeviation_constructor():
    sq_dev = SquaredDeviation(d_ref=2, priority=100)
    assert sq_dev.parameters == [2.0]
    assert sq_dev.name == "Squared Deviation"
    assert sq_dev.d_ref == 2.0
    assert sq_dev.priority == 100.0


def test_SquaredDeviation_compute_objective():
    dose = np.array([1, 2, 3])
    sq_dev = SquaredDeviation(d_ref=2.0)
    assert sq_dev.compute_objective(dose) == 2 / 3


def test_SquaredDeviation_compute_gradient():
    dose = np.array([1, 2, 3])
    sq_dev = SquaredDeviation(d_ref=2.0)
    grad_expected = 2 / 3 * np.array([-1, 0, 1])
    assert np.all(sq_dev.compute_gradient(dose) == grad_expected)


def test_SquaredOverdosing_constructor():
    sq_over = SquaredOverdosing(d_max=2, priority=100)
    assert sq_over.parameter_names == ["d_max"]
    # assert sq_over.parameter_types == ["dose"]
    assert sq_over.parameters == [2.0]
    assert sq_over.d_max == 2.0
    assert sq_over.priority == 100.0


def test_SquaredOverdosing_compute_objective():
    sq_over = SquaredOverdosing(d_max=2.0)
    dose = np.array([1, 2, 3])
    assert sq_over.compute_objective(dose) == 1 / 3


def test_SquaredOverdosing_compute_gradient():
    dose = np.array([1, 2, 3])
    sq_over = SquaredOverdosing(d_max=2)
    grad_expected = 2 / 3 * np.array([0, 0, 1])
    assert np.all(sq_over.compute_gradient(dose) == grad_expected)


def test_SquaredUnderdosing_constructor():
    sq_under = SquaredUnderdosing(d_min=2, priority=100)
    assert sq_under.name == "Squared Underdosing"
    assert sq_under.parameter_names == ["d_min"]
    # assert sq_under.parameter_types == ["dose"]
    assert sq_under.parameters == [2.0]
    assert sq_under.d_min == 2.0
    assert sq_under.priority == 100.0


def test_SquaredUnderdosing_compute_objective():
    sq_under = SquaredUnderdosing(d_min=2.0)
    dose = np.array([1, 2, 3])
    assert sq_under.compute_objective(dose) == 1 / 3


def test_SquaredUnderdosing_compute_gradient():
    dose = np.array([1, 2, 3])
    sq_under = SquaredUnderdosing(d_min=2.0)
    grad_expected = 2 / 3 * np.array([-1, 0, 0])
    assert np.all(sq_under.compute_gradient(dose) == grad_expected)


def test_EUD_constructor():
    eud = EUD(k=3, eud_ref=0.0, priority=100)
    assert eud.name == "EUD"
    assert eud.parameter_names == ["eud_ref", "k"]
    assert eud.parameter_types == ["reference", "numeric"]
    assert eud.parameters == [0.0, 3.0]
    assert eud.priority == 100.0


def test_EUD_compute_objective():
    eud = EUD(k=3, EUD_ref=0.0)
    dose = np.array([1, 2, 3])
    assert (eud.compute_objective(dose) - (1 / 3 * (1 + 2 ** (1 / 3) + 3 ** (1 / 3))) ** 6) < 1e-10


def test_EUD_compute_gradient():
    eud_obj = EUD(k=3, EUD_ref=0.0)
    dose = np.array([1, 2, 3])
    d_eud = (1 + 2 ** (1 / 3) + 3 ** (1 / 3)) ** 2 * np.array([1, 2, 3]) ** (-2 / 3) * 1 / 3**3
    eud = (1 / 3 * (1 + 2 ** (1 / 3) + 3 ** (1 / 3))) ** 3
    grad_expected = 2 * (eud - 0) * d_eud
    assert np.all((eud_obj.compute_gradient(dose) - grad_expected) < 1e-10)


def test_MeanDose_constructor():
    mean_dose = MeanDose(d_ref=2, priority=100)
    assert mean_dose.name == "Mean Dose"
    assert mean_dose.parameter_names == ["d_ref"]
    assert mean_dose.parameter_types == ["reference"]
    assert mean_dose.parameters == [2.0]
    assert mean_dose.d_ref == 2.0
    assert mean_dose.priority == 100.0


def test_MeanDose_compute_objective():
    mean_dose = MeanDose(d_ref=2.0)
    dose = np.array([1, 2, 3])
    assert mean_dose.compute_objective(dose) == 0


def test_MeanDose_compute_gradient():
    mean_dose = MeanDose(d_ref=2.0)
    dose = np.array([1, 2, 3])
    grad_expected = np.zeros(3)
    assert np.all(mean_dose.compute_gradient(dose) == grad_expected)


def test_MinDVH_constructor():
    min_dvh = MinDVH(d=2, v_min=3, priority=100)
    assert min_dvh.name == "Min DVH"
    assert min_dvh.parameter_names == ["d", "v_min"]
    assert min_dvh.parameter_types == ["reference", "relative_volume"]
    assert min_dvh.d == 2.0
    assert min_dvh.v_min == 3.0
    assert min_dvh.priority == 100.0
    assert min_dvh.parameters == [2.0, 3.0]


def test_MinDVH_compute_objective():
    min_dvh = MinDVH(d=30.0, v_min=95)
    dose = np.ones(100)
    dose_2 = np.ones(100) * 50
    assert min_dvh.compute_objective(dose) == 841
    assert min_dvh.compute_objective(dose_2) == 0


def test_MinDVH_compute_gradient():
    min_dvh = MinDVH(d=30.0, v_min=95)
    dose = np.ones(100)
    dose_2 = np.ones(100) * 50
    grad_expected = np.ones(100) * -0.58
    grad_expected2 = np.zeros(100)
    assert np.all(min_dvh.compute_gradient(dose) == grad_expected)
    assert np.all(min_dvh.compute_gradient(dose_2) == grad_expected2)


def test_MaxDVH_constructor():
    max_dvh = MaxDVH(d=30.0, v_max=50, priority=100)
    assert max_dvh.name == "Max DVH"
    assert max_dvh.parameter_names == ["d", "v_max"]
    assert max_dvh.parameter_types == ["reference", "relative_volume"]
    assert max_dvh.d == 30.0
    assert max_dvh.v_max == 50.0
    assert max_dvh.priority == 100.0
    assert max_dvh.parameters == [30.0, 50.0]


def test_MaxDVH_compute_objective():
    max_dvh = MaxDVH(d=30.0, v_max=50)
    dose = np.ones(100)
    dose_2 = np.ones(100) * 50
    assert max_dvh.compute_objective(dose) == 0
    assert max_dvh.compute_objective(dose_2) == 400


def test_MaxDVH_compute_gradient():
    max_dvh = MaxDVH(d=30.0, v_max=50)
    dose = np.ones(100)
    dose_2 = np.ones(100) * 50
    grad_expected = np.zeros(100)
    grad_expected2 = np.ones(100) * 0.4
    assert np.all(max_dvh.compute_gradient(dose) == grad_expected)
    assert np.all(max_dvh.compute_gradient(dose_2) == grad_expected2)



import numpy as np

from scipy.sparse import hstack
from ._dij import Dij


def compose_beam_dijs(dijs: list[Dij]) -> Dij:
    """Compose multiple Dij objects into a single Dij object.

    Parameters
    ----------
    dijs : list[Dij]
        List of Dij objects to be composed.

    Returns
    -------
    Dij
        Composed Dij object.
    """

    qs = dijs[0].quantities

    num_of_beams = 0
    beam_num = np.array([], dtype=int)
    ray_num = np.array([], dtype=int)
    bixel_num = np.array([], dtype=int)

    for i, dij in enumerate(dijs):
        beam_num = np.append(beam_num, dij.beam_num + num_of_beams)
        ray_num = np.append(ray_num, dij.ray_num)
        bixel_num = np.append(bixel_num, dij.bixel_num)
        num_of_beams = num_of_beams + dij.num_of_beams

    matrices = {}
    for q in qs:
        tmp_matrices = [getattr(dij, q) for dij in dijs]

        new_matrix = np.empty_like(tmp_matrices[0], dtype=object)

        for i, scen_matrix in enumerate(tmp_matrices[0].flat):
            if scen_matrix is not None:
                new_matrix.flat[i] = hstack([tmp_matrix.flat[i] for tmp_matrix in tmp_matrices])

        matrices.update({q: new_matrix})

    return Dij(
        ct_grid=dijs[0].ct_grid,
        dose_grid=dijs[0].dose_grid,
        matrices=matrices,
        num_of_beams=num_of_beams,
        beam_num=beam_num,
        ray_num=ray_num,
        bixel_num=bixel_num,
        **matrices,
    )



"""Solvers module providing different solvers for pyRadPlan."""

from ._factory import register_solver, get_available_solvers, get_solver

try:
    from ._ipopt import OptimizerIpopt

    register_solver(OptimizerIpopt)

except ImportError:
    OptimizerIpopt = None

from ._base_solvers import SolverBase, NonLinearOptimizer
from ._scipy_solver import OptimizerSciPy

register_solver(OptimizerSciPy)


__all__ = [
    "OptimizerIpopt",
    "OptimizerSciPy",
    "SolverBase",
    "NonLinearOptimizer",
    "register_solver",
    "get_available_solvers",
    "get_solver",
]



import pytest
import SimpleITK as sitk
import os
import numpy as np
from pathlib import Path
from pydantic import ValidationError

from pyRadPlan.ct import CT, create_ct, validate_ct, ct_from_file


@pytest.fixture
def sample_image(tmp_path):
    """Create a sample SimpleITK image for testing."""
    sample_array = np.random.rand(50, 100, 100) * 1000  # Random HU values
    image = sitk.GetImageFromArray(sample_array)
    image.SetOrigin((0, 0, 0))
    image.SetSpacing((1, 1, 2))
    image.SetDirection((1, 0, 0, 0, 1, 0, 0, 0, 1))

    """Save it for testing ct_from_file method"""
    writer = sitk.ImageFileWriter()
    writer.SetUseCompression(True)
    writer.SetImageIO("NiftiImageIO")

    file_path = os.path.join(tmp_path, "DummyCT.nii.gz")

    writer.SetFileName(file_path)
    writer.Execute(image)

    return image, file_path


def test_ct_creation(sample_image):
    """Test creation of a CT object from a SimpleITK image."""
    ct = create_ct(cube_hu=sample_image[0])
    assert isinstance(ct, CT)
    assert ct.size == (100, 100, 50)  # reverse order of axes
    assert ct.resolution == {"x": 1, "y": 1, "z": 2}
    assert ct.origin == (0, 0, 0)
    assert ct.direction == (1, 0, 0, 0, 1, 0, 0, 0, 1)


def test_ct_from_dict(sample_image):
    """Test creation of a CT object from a dictionary."""
    ct_dict = {
        "cube_hu": sample_image[0],
        "origin": (1, 2, 3),
        "direction": (1, 0, 0, 0, 1, 0, 0, 0, 1),
        "resolution": {"x": 2, "y": 2, "z": 4},
    }
    ct = create_ct(ct_dict)
    assert ct.origin == (1, 2, 3)
    assert ct.resolution == {"x": 2, "y": 2, "z": 4}


def test_ct_to_matrad_dict(sample_image):
    """Test conversion of CT object to matRad dictionary format."""
    ct = create_ct(cube_hu=sample_image[0])
    matrad_dict = ct.to_matrad()
    assert "cubeHU" in matrad_dict
    assert "cubeDim" in matrad_dict
    assert matrad_dict["cubeDim"].tolist() == [100.0, 100.0, 50.0]  # reverse order of axes
    assert matrad_dict["cubeHU"].shape == (1,)
    assert matrad_dict["cubeHU"][0].shape == (100, 100, 50)  # reverse order of axes


def test_ct_from_file(sample_image):
    """Test creation of a CT object from a .nii.gz file."""
    sample_file = Path(sample_image[1])
    if sample_file.exists():
        ct = ct_from_file(sample_file)
        assert isinstance(ct, CT)
    else:
        pytest.skip("Sample DICOM file not available")


def test_validate_ct(sample_image):
    """Test validation of CT objects."""
    ct = create_ct(cube_hu=sample_image[0])
    validated_ct = validate_ct(ct)
    assert ct is validated_ct

    validated_ct = validate_ct(cube_hu=sample_image[0])
    assert isinstance(validated_ct, CT)


def test_invalid_input():
    """Test handling of invalid input."""
    with pytest.raises(ValueError):
        create_ct(cube_hu="invalid input")


@pytest.mark.parametrize(
    "size,spacing",
    [((5, 5, 5), (2, 2, 2)), ((20, 30, 40), (0.5, 0.5, 1)), ((100, 100, 100), (0.1, 0.1, 0.1))],
)
def test_various_image_sizes(size, spacing):
    """Test CT creation with various image sizes and spacings."""
    # Numpy axis order is z,y,x and SimpleITK axis order is x,y,z. Need to flip size before assertion
    sample_array = np.random.rand(*size) * 1000
    image = sitk.GetImageFromArray(sample_array)
    image.SetSpacing(spacing)
    ct = create_ct(cube_hu=image)
    assert ct.size == size[::-1]
    assert ct.resolution == {"x": spacing[0], "y": spacing[1], "z": spacing[2]}


def test_ct_property_change(sample_image):
    """Test changing properties of an existing CT object."""
    ct = create_ct(cube_hu=sample_image[0])
    new_origin = (5, 5, 1)
    ct.cube_hu.SetOrigin(new_origin)
    assert ct.origin == new_origin


def test_check_dict_convert_sitk_image():
    """Test the check_dict_convert_sitk_image method."""
    sample_array = np.random.rand(100, 200, 200) * 1000
    ct_dict = {
        "cube_hu": sample_array,
        "origin": (1, 2, 3),
        "direction": (1, 0, 0, 0, 1, 0, 0, 0, 1),
        "resolution": {"x": 2, "y": 2, "z": 4},
    }

    ct = CT.model_validate(ct_dict)
    assert isinstance(ct.cube_hu, sitk.Image)
    assert ct.origin == (1, 2, 3)
    assert ct.resolution == {"x": 2, "y": 2, "z": 4}


def test_check_multiple_scenarios_from_numpy():
    """Test the conversion to 4D image."""
    sample_array = np.ndarray((2,), dtype=object)
    sample_array[0] = np.random.rand(100, 200, 200) * 1000
    sample_array[1] = np.random.rand(100, 200, 200) * 1000
    ct_dict = {
        "cube_hu": sample_array,
        "origin": (1, 2, 3, 0),
        "direction": (1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1),
        "resolution": {"x": 2, "y": 2, "z": 4},
    }

    ct = CT.model_validate(ct_dict)
    assert isinstance(ct.cube_hu, sitk.Image)
    assert ct.origin == (1, 2, 3)
    assert ct.resolution == {"x": 2, "y": 2, "z": 4}


def test_missing_hu_cube():
    """Test handling of missing HU cube."""
    with pytest.raises(ValidationError):
        CT.model_validate({})


# WET
@pytest.fixture
def sample_ct(sample_image):
    """Create a sample CT object for testing."""
    return create_ct(cube_hu=sample_image[0])


def test_compute_wet_valid_hlut(sample_ct):
    """Test compute_wet with a valid HLUT."""
    hlut = np.array(
        [[-1000, 0, 3000], [0, 1, 2]]  # HU values  # Corresponding WET values
    ).transpose()
    wet_image = sample_ct.compute_wet(hlut)
    assert isinstance(wet_image, sitk.Image)
    assert wet_image.GetSize() == sample_ct.size
    assert wet_image.GetSpacing() == (
        sample_ct.resolution["x"],
        sample_ct.resolution["y"],
        sample_ct.resolution["z"],
    )
    assert wet_image.GetOrigin() == sample_ct.origin
    assert wet_image.GetDirection() == sample_ct.direction


def test_compute_wet_invalid_hlut(sample_ct):
    """Test compute_wet with an invalid HLUT."""
    hlut = np.array([[-1000, 0, 1000]])  # Invalid HLUT with only one row
    with pytest.raises(ValueError):
        sample_ct.compute_wet(hlut)


def test_compute_wet_edge_cases(sample_ct):
    """Test compute_wet with edge cases in HLUT."""
    hlut = np.array(
        [[-1000, 0, 1000], [0, 1, 2]]  # HU values  # Corresponding WET values
    ).transpose()
    # Test with minimum HU value
    sample_ct.cube_hu = sitk.GetImageFromArray(np.full((50, 100, 100), -1000))
    wet_image = sample_ct.compute_wet(hlut)
    assert np.all(sitk.GetArrayFromImage(wet_image) == 0)

    # Test with maximum HU value
    sample_ct.cube_hu = sitk.GetImageFromArray(np.full((50, 100, 100), 1000))
    wet_image = sample_ct.compute_wet(hlut)
    assert np.all(sitk.GetArrayFromImage(wet_image) == 2)


def test_hlut_unsorted(sample_ct):
    """Test compute_wet with an unsorted HLUT."""
    hlut_unsorted = np.array(
        [[0, -1000, 3000], [1, 0, 3]]  # HU values  # Corresponding WET values
    ).transpose()

    hlut_sorted = np.array(
        [[-1000, 0, 3000], [0, 1, 3]]  # HU values  # Corresponding WET values
    ).transpose()

    wet_image_unsorted = sample_ct.compute_wet(hlut_unsorted)
    wet_image_sorted = sample_ct.compute_wet(hlut_sorted)
    assert np.all(
        sitk.GetArrayViewFromImage(wet_image_unsorted)
        == sitk.GetArrayViewFromImage(wet_image_sorted)
    )



"""Minimum DVH objective."""

from typing import Annotated
from pydantic import Field

from numba import njit
from numpy import logical_or, quantile, sort

from ._objective import Objective, ParameterMetadata


class MinDVH(Objective):
    """
    Minimum DVH objective.

    Attributes
    ----------
    d : float
        dose point
    v_min : float
        min. relative volume [%]
    """

    name = "Min DVH"

    d: Annotated[float, Field(default=30.0, ge=0.0), ParameterMetadata(kind="reference")]
    v_min: Annotated[
        float, Field(default=95.0, ge=0.0, le=100.0), ParameterMetadata(kind="relative_volume")
    ]

    def compute_objective(self, values):
        return _compute_objective(values, self.d, self.v_min)

    def compute_gradient(self, values):
        return _compute_gradient(values, self.d, self.v_min)


@njit
def _compute_objective(dose, d, v_min):
    deviation = dose - d
    dose_quantile = quantile(sort(dose)[::-1], v_min / 100.0)
    mask = logical_or(dose > d, dose < dose_quantile)
    deviation[mask] = 0

    return (deviation @ deviation) / len(dose)


@njit
def _compute_gradient(dose, d, v_min):
    deviation = dose - d
    dose_quantile = quantile(sort(dose)[::-1], v_min / 100.0)
    mask = logical_or(dose > d, dose < dose_quantile)
    deviation[mask] = 0
    return 2.0 * deviation / len(dose)



"""Squared Underdosing Objective."""

from typing import Annotated
from pydantic import Field

from numba import njit
from numpy import clip

from ._objective import Objective, ParameterMetadata

# %% Class definition


class SquaredUnderdosing(Objective):
    """
    Squared Underdosing (piece-wise negative least-squares) objective.

    Attributes
    ----------
    d_min : float
        minimum dose value (below which we penalize)
    """

    name = "Squared Underdosing"

    d_min: Annotated[float, Field(default=60.0, ge=0.0), ParameterMetadata(kind="reference")]

    def compute_objective(self, values):
        return _compute_objective(values, self.d_min)

    def compute_gradient(self, values):
        return _compute_gradient(values, self.d_min)


@njit
def _compute_objective(dose, d_min):
    underdose = clip(dose - d_min, a_min=None, a_max=0)

    return (underdose @ underdose) / len(dose)


@njit
def _compute_gradient(dose, d_min):
    underdose = clip(dose - d_min, a_min=None, a_max=0)
    return 2.0 * underdose / len(underdose)



import numpy as np
import SimpleITK as sitk
from pyRadPlan.scenarios import ScenarioModel, NominalScenario, validate_scenario_model
from pyRadPlan.ct import CT
import pytest


@pytest.fixture
def sample_model_dict_matrad():
    model_dict = {
        "model": "nomScen",
        "rangeRelSD": 5.0,
        "rangeAbsSD": 2.5,
        "shiftSD": (3.0, 3.0, 3.0),
        # TODO!: MatRad export has e.g. (1, 0.5) as indexing.
        # The Model does not yet respect that (converting 1 to 0)
        # because it cant distinguish whether its from the user or matRad. (adding '1' or keeping '0')
        "ctScenProb": [(1, 0.5)],
        "wcSigma": 5.0,
    }
    return model_dict


def helper_mvar_gauss(scenarioModel):
    # Computes multivariate Gaussian probability for scenario model
    # Can be used to test correct scenario probabilities. Also considers used
    # ct phases for probability computation (uncorrelated)

    sigma = np.diag(
        np.hstack(
            (
                scenarioModel.shift_sd,
                scenarioModel.range_abs_sd,
                scenarioModel.range_rel_sd / 100.0,
            )
        )
        ** 2
    )
    d = sigma.shape[0]
    cs = np.linalg.cholesky(sigma)

    tmp_standardized = np.linalg.solve(cs, scenarioModel.scen_for_prob[:, 1:].T).T
    p = (
        (2 * np.pi) ** (-d / 2)
        * np.exp(-0.5 * np.sum(tmp_standardized**2, axis=1))
        / np.prod(np.diag(cs))
    )

    # Now multiply with the phase probability
    tmp_phase_prob = np.array(
        [
            [x[1] for x in scenarioModel.ct_scen_prob if x[0] == phase]
            for phase in scenarioModel.scen_for_prob[:, 0]
        ]
    )
    p = p * tmp_phase_prob

    return p


@pytest.fixture
def sample_ct():
    """Create a sample SimpleITK image for testing."""
    sample_array = np.random.rand(50, 100, 100) * 1000  # Random HU values
    image = sitk.GetImageFromArray(sample_array)
    image.SetOrigin((0, 0, 0))
    image.SetSpacing((1, 1, 2))
    image.SetDirection((1, 0, 0, 0, 1, 0, 0, 0, 1))

    ct = CT(cube_hu=image)
    return ct


# Test Abstract Scenario Model
def test_scenario_model():
    with pytest.raises(TypeError):
        ScenarioModel()


def test_nominal_scenario_construct():
    scenario = NominalScenario()
    assert isinstance(scenario, ScenarioModel)
    assert isinstance(scenario, NominalScenario)
    assert scenario.name == "Nominal Scenario"
    assert scenario.short_name == "nomScen"
    assert scenario.ct_scen_prob == [(0, 1.0)]
    assert scenario.num_of_ct_scen == 1
    assert scenario.tot_num_scen == 1
    assert scenario.tot_num_shift_scen == 1
    assert scenario.tot_num_range_scen == 1
    assert scenario.rel_range_shift == 0.0
    assert scenario.abs_range_shift == 0.0
    assert np.array_equal(scenario.iso_shift, np.array([[0.0, 0.0, 0.0]]))
    assert scenario.max_abs_range_shift == 0.0
    assert scenario.max_rel_range_shift == 0.0
    assert np.array_equal(scenario.scen_mask, np.ones((1, 1, 1)).astype(bool))
    assert np.array_equal(scenario.linear_mask, np.array([[0, 0, 0]]))
    assert np.allclose(scenario.scen_prob, helper_mvar_gauss(scenario))
    assert np.array_equal(
        scenario.scen_for_prob,
        np.hstack((np.zeros((1, 1), dtype=float), np.zeros((1, 5), dtype=float))),
    )
    assert np.array_equal(scenario.scen_weight, np.ones((1,), dtype=float))


def test_nominal_scenario_construct_with_ct(sample_ct):
    scenario = NominalScenario(sample_ct)
    assert isinstance(scenario, ScenarioModel)
    assert isinstance(scenario, NominalScenario)
    assert scenario.name == "Nominal Scenario"
    assert scenario.short_name == "nomScen"
    assert scenario.short_name == "nomScen"
    assert scenario.ct_scen_prob == [(0, 1.0)]
    assert scenario.num_of_ct_scen == 1
    assert scenario.tot_num_scen == 1
    assert scenario.tot_num_shift_scen == 1
    assert scenario.tot_num_range_scen == 1
    assert scenario.rel_range_shift == 0.0
    assert scenario.abs_range_shift == 0.0
    assert np.array_equal(scenario.iso_shift, np.array([[0.0, 0.0, 0.0]]))
    assert scenario.max_abs_range_shift == 0.0
    assert scenario.max_rel_range_shift == 0.0
    assert np.array_equal(scenario.scen_mask, np.ones((1, 1, 1)).astype(bool))
    assert np.array_equal(scenario.linear_mask, np.array([[0, 0, 0]]))
    assert np.allclose(scenario.scen_prob, helper_mvar_gauss(scenario))
    assert np.array_equal(
        scenario.scen_for_prob,
        np.hstack((np.zeros((1, 1), dtype=float), np.zeros((1, 5), dtype=float))),
    )
    assert np.array_equal(scenario.scen_weight, np.ones((1,), dtype=float))


def test_nominal_scenario_list():
    scenario = NominalScenario()
    scenario.list_all_scenarios()


def test_nominal_scenario_creation_by_name():
    scenario = validate_scenario_model("nomScen")
    assert isinstance(scenario, ScenarioModel)
    assert isinstance(scenario, NominalScenario)

    with pytest.raises(ValueError):
        validate_scenario_model("unknown")

    with pytest.raises(NotImplementedError):
        validate_scenario_model("wcScen")

    with pytest.raises(NotImplementedError):
        validate_scenario_model("impScen")

    with pytest.raises(NotImplementedError):
        validate_scenario_model("rndScen")


def test_nominal_scenario_creation_by_dict():
    model_dict = {
        "model": "nomScen",
        "range_rel_sd": 5.0,
        "range_abs_sd": 2.5,
        "shift_sd": (3.0, 3.0, 3.0),
        "ct_scen_prob": [(0, 0.5)],
        "wc_sigma": 5.0,
    }

    scenario = validate_scenario_model(model_dict)

    assert isinstance(scenario, ScenarioModel)
    assert isinstance(scenario, NominalScenario)
    assert scenario.range_rel_sd == 5.0
    assert scenario.range_abs_sd == 2.5
    assert np.array_equal(scenario.shift_sd, (3.0, 3.0, 3.0))
    assert scenario.ct_scen_prob == [(0, 0.5)]
    assert scenario.wc_sigma == 5.0

    return_dict = scenario.model_dump()

    model_dict.pop("model")

    for key in model_dict.keys():
        assert key in return_dict
        assert return_dict[key] == model_dict[key]


def test_nominal_scenario_creation_by_matrad_dict(sample_model_dict_matrad):
    scenario = validate_scenario_model(sample_model_dict_matrad)

    assert isinstance(scenario, ScenarioModel)
    assert isinstance(scenario, NominalScenario)
    assert scenario.range_rel_sd == 5.0
    assert scenario.range_abs_sd == 2.5
    assert np.array_equal(scenario.shift_sd, [3.0, 3.0, 3.0])
    assert scenario.ct_scen_prob == [(0, 0.5)]
    assert scenario.wc_sigma == 5.0

    return_dict = scenario.to_matrad()

    sample_model_dict_matrad.pop("model")

    for key in sample_model_dict_matrad.keys():
        assert key in return_dict
        assert return_dict[key] == sample_model_dict_matrad[key]


def test_nominal_scenario_extract_single_scenario():
    scenario = NominalScenario()
    assert isinstance(scenario.extract_single_scenario(0), ScenarioModel)
    assert isinstance(scenario.extract_single_scenario(0), NominalScenario)

    with pytest.raises(NotImplementedError):
        scenario.extract_single_scenario(1)



import os
import pytest

import SimpleITK as sitk
import numpy as np

import pyRadPlan.io.matfile as matfile
from pyRadPlan.ct import create_ct


from pyRadPlan.ct import CT
from pyRadPlan.cst import (
    StructureSet,
    create_cst,
    validate_cst,
    create_voi,
    ExternalVOI,
    HelperVOI,
    Target,
    OAR,
)


def test_cst_from_matrad_mat_file(matrad_import):
    ct = create_ct(matrad_import["ct"])
    cst = create_cst(matrad_import["cst"], ct=ct)
    assert isinstance(cst, StructureSet)

    cst = validate_cst(matrad_import["cst"], ct=ct)
    assert isinstance(cst, StructureSet)

    with pytest.raises(ValueError):
        cst = create_cst(matrad_import["cst"])

    with pytest.raises(ValueError):
        cst = validate_cst(matrad_import["cst"])


def test_cst_to_matrad(matrad_import, tmpdir):
    ct = create_ct(matrad_import["ct"])
    cst = create_cst(matrad_import["cst"], ct=ct)

    matrad_list = cst.to_matrad()
    assert isinstance(matrad_list, list)

    tmp_mat_path = os.path.join(tmpdir, "test_cst.mat")
    matfile.save(tmp_mat_path, {"cst": matrad_list})
    assert os.path.exists(tmp_mat_path)

    tmp = matfile.load(tmp_mat_path)

    assert isinstance(tmp, dict)
    assert isinstance(tmp["cst"], list)


def test_cst_target_voxels(generic_input_3d):
    name_3d, ct_3d, mask_3d, _, _ = generic_input_3d

    mask_3d_2 = mask_3d.copy()
    mask_3d_2.fill(0)
    mask_3d_2[0, 0, 0] = 1

    mask_3d_3 = mask_3d.copy()
    mask_3d_3.fill(0)
    mask_3d_3[1, 1, 1] = 1

    mask_3d_4 = mask_3d.copy()
    mask_3d_4.fill(0)
    mask_3d_4[0, 1, 0] = 1

    voi_3d = create_voi(voi_type="TARGET", name=name_3d, ct_image=ct_3d, mask=mask_3d)
    voi_3d_2 = create_voi(voi_type="TARGET", name=name_3d, ct_image=ct_3d, mask=mask_3d_2)
    voi_3d_3 = create_voi(voi_type="OAR", name=name_3d, ct_image=ct_3d, mask=mask_3d_3)
    voi_3d_4 = create_voi(voi_type="EXTERNAL", name=name_3d, ct_image=ct_3d, mask=mask_3d_4)

    cst = StructureSet(vois=[voi_3d, voi_3d_2, voi_3d_3, voi_3d_4], ct_image=ct_3d)

    index_union = cst.target_union_voxels()
    assert (index_union == cst.target_union_voxels(order="sitk")).all()
    index_union_np = cst.target_union_voxels(order="numpy")
    mask_union = cst.target_union_mask()

    assert (index_union == np.array([0, 5000])).all()
    assert (index_union_np == np.array([0, 1])).all()
    assert (sitk.GetArrayViewFromImage(mask_union).ravel(order="F")[index_union] == 1).all()
    assert (sitk.GetArrayViewFromImage(mask_union).ravel(order="C")[index_union_np] == 1).all()


def test_cst_patient_voxels(generic_input_3d):
    name_3d, ct_3d, mask_3d, _, _ = generic_input_3d

    mask_3d_3 = mask_3d.copy()
    mask_3d_3.fill(0)
    mask_3d_3[0, 0, 0] = 1

    mask_3d_4 = mask_3d.copy()
    mask_3d_4.fill(0)
    mask_3d_4[-1, -1, -1] = 1

    voi_3d = create_voi(voi_type="TARGET", name=name_3d, ct_image=ct_3d, mask=mask_3d)
    voi_3d_3 = create_voi(voi_type="OAR", name=name_3d, ct_image=ct_3d, mask=mask_3d_3)
    voi_3d_4 = create_voi(voi_type="OAR", name=name_3d, ct_image=ct_3d, mask=mask_3d_4)

    cst = StructureSet(vois=[voi_3d, voi_3d_3, voi_3d_4], ct_image=ct_3d)

    index_union = cst.patient_voxels()
    assert (index_union == cst.patient_voxels(order="sitk")).all()
    index_union_np = cst.patient_voxels(order="numpy")
    mask_union = cst.patient_mask()

    assert (index_union == np.array([0, 5000, mask_3d_3.size - 1])).all()
    assert (index_union_np == np.array([0, 1, mask_3d_3.size - 1])).all()
    assert (sitk.GetArrayViewFromImage(mask_union).ravel(order="F")[index_union] == 1).all()
    assert (sitk.GetArrayViewFromImage(mask_union).ravel(order="C")[index_union_np] == 1).all()

    voi_3d_4 = create_voi(voi_type="EXTERNAL", name=name_3d, ct_image=ct_3d, mask=mask_3d_4)
    cst.vois[-1] = voi_3d_4
    index_union = cst.patient_voxels()
    assert (index_union == cst.patient_voxels(order="sitk")).all()
    index_union_np = cst.patient_voxels(order="numpy")
    mask_union = cst.patient_mask()

    assert (index_union == np.array([mask_3d_4.size - 1])).all()
    assert (index_union_np == np.array([mask_3d_4.size - 1])).all()


def test_target_center_of_mass():
    ct = create_ct(cube_hu=sitk.Image(10, 10, 10, sitk.sitkInt16))
    mask = np.zeros((10, 10, 10), dtype=np.uint8)
    mask[0, 0, 0] = 1
    mask[1, 1, 1] = 1
    mask[2, 2, 2] = 1
    mask[3, 3, 3] = 1

    voi = create_voi(voi_type="TARGET", name="test", ct_image=ct, mask=mask)
    cst = StructureSet(vois=[voi], ct_image=ct)

    com = cst.target_center_of_mass()
    assert np.allclose(com, np.array([1.5, 1.5, 1.5]))

    image4d = sitk.JoinSeries([sitk.Image(10, 10, 10, sitk.sitkInt16) for _ in range(3)])
    ct = create_ct(cube_hu=image4d)
    mask = np.zeros((3, 10, 10, 10), dtype=np.uint8)
    mask[0, 0, 0, 0] = 1
    mask[0, 1, 1, 1] = 1
    mask[0, 2, 2, 2] = 1
    mask[0, 3, 3, 3] = 1

    voi = create_voi(voi_type="TARGET", name="test", ct_image=ct, mask=mask)
    cst = StructureSet(vois=[voi], ct_image=ct)

    com = cst.target_center_of_mass()
    assert np.allclose(com, np.array([1.5, 1.5, 1.5]))


@pytest.fixture
def generic_ct():
    # Create a simple 3D CT image
    ct_array = np.zeros((10, 10, 10), dtype=np.float32)
    ct_image = sitk.GetImageFromArray(ct_array)
    ct_image.SetSpacing((1.0, 1.0, 1.0))
    ct_image.SetOrigin((0.0, 0.0, 0.0))
    ct_image.SetDirection((1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0))
    return CT(cube_hu=ct_image)


@pytest.fixture
def generic_vois(generic_ct):
    # Create simple VOIs with different overlap priorities
    mask1 = np.zeros((10, 10, 10), dtype=np.uint8)
    mask1[4:5, 4:5, 4:5] = 1
    mask_image1 = sitk.GetImageFromArray(mask1)
    mask_image1.CopyInformation(generic_ct.cube_hu)

    mask2 = np.zeros((10, 10, 10), dtype=np.uint8)
    mask2[3:6, 3:6, 3:6] = 1
    mask_image2 = sitk.GetImageFromArray(mask2)
    mask_image2.CopyInformation(generic_ct.cube_hu)

    mask3 = np.zeros((10, 10, 10), dtype=np.uint8)
    mask3[4:7, 4:7, 4:7] = 1
    mask_image3 = sitk.GetImageFromArray(mask3)
    mask_image3.CopyInformation(generic_ct.cube_hu)

    mask4 = np.zeros((10, 10, 10), dtype=np.uint8)
    mask4[1:8, 1:8, 1:8] = 1
    mask_image4 = sitk.GetImageFromArray(mask4)
    mask_image4.CopyInformation(generic_ct.cube_hu)

    mask5 = np.zeros((10, 10, 10), dtype=np.uint8)
    mask5[2:7, 2:7, 2:7] = 1
    mask_image5 = sitk.GetImageFromArray(mask5)
    mask_image5.CopyInformation(generic_ct.cube_hu)

    voi1 = Target(name="CTV", mask=mask_image1, ct_image=generic_ct, overlap_priority=1)
    voi2 = Target(name="PTV", mask=mask_image2, ct_image=generic_ct, overlap_priority=2)
    voi3 = OAR(name="OAR", mask=mask_image3, ct_image=generic_ct)
    voi4 = ExternalVOI(name="BODY", mask=mask_image4, ct_image=generic_ct)
    voi5 = HelperVOI(name="HELPER", mask=mask_image5, ct_image=generic_ct)

    return [voi1, voi2, voi3, voi4, voi5]


def test_apply_overlap_priorities(generic_ct, generic_vois):
    # Create a StructureSet with the VOIs
    structure_set = StructureSet(ct_image=generic_ct, vois=generic_vois)

    # Apply overlap priorities
    structure_set_overlap = structure_set.apply_overlap_priorities()

    voi_mask = [None] * len(generic_vois)
    voi_mask_overlapped = [None] * len(generic_vois)
    p = [None] * len(generic_vois)

    for i in range(len(generic_vois)):
        voi_mask[i] = sitk.GetArrayViewFromImage(structure_set.vois[i].mask)
        voi_mask_overlapped[i] = sitk.GetArrayViewFromImage(structure_set_overlap.vois[i].mask)
        p[i] = structure_set.vois[i].overlap_priority

    expected_overlap_list = np.argsort(p)

    ol_mask = np.zeros(voi_mask[0].shape, dtype=bool)

    for expected in expected_overlap_list:
        # the mask that the current voi should have
        assert (
            voi_mask_overlapped[expected][np.logical_and(voi_mask[expected] > 0, ~ol_mask)]
        ).all()

        # Accumulate the ol mask
        ol_mask = ol_mask | voi_mask[expected] > 0

        # we currently should be zero where overlapped
        assert not (voi_mask_overlapped[expected][ol_mask] == 0).all()
        # the accumulated ol mask


def test_apply_overlap_priorities_same_priority(generic_ct, generic_vois):
    for v in generic_vois:
        v.overlap_priority = 1

    # Create a StructureSet with the VOIs
    structure_set = StructureSet(ct_image=generic_ct, vois=generic_vois)

    # Apply overlap priorities
    structure_set_overlap = structure_set.apply_overlap_priorities()

    # Check the resulting masks
    for i in range(len(generic_vois)):
        voi_mask = sitk.GetArrayViewFromImage(structure_set.vois[i].mask)
        voi_mask_overlapped = sitk.GetArrayViewFromImage(structure_set_overlap.vois[i].mask)

        assert np.isclose(voi_mask, voi_mask_overlapped).all()



import warnings
import logging
from typing import Union, Type
from pyRadPlan.plan import validate_pln, Plan
from pyRadPlan.dose.engines import DoseEngineBase

DOSE_ENGINES = {}

logger = logging.getLogger(__name__)


def register_engine(engine_cls: Type[DoseEngineBase]) -> None:
    """
    Register a new engine.

    Parameters
    ----------
    engine_cls : type
        A Dose Engine class.
    """
    if not issubclass(engine_cls, DoseEngineBase):
        raise ValueError("Engine must be a subclass of DoseEngineBase.")

    if engine_cls.short_name is None:
        raise ValueError("Engine must have a 'short_name' attribute.")

    if engine_cls.name is None:
        raise ValueError("Engine must have a 'name' attribute.")

    if engine_cls.possible_radiation_modes is None:
        raise ValueError("Engine must have a 'possible_radiation_modes' attribute.")

    engine_name = engine_cls.short_name
    if engine_name in DOSE_ENGINES:
        warnings.warn(f"Engine '{engine_name}' is already registered.")
    else:
        DOSE_ENGINES[engine_name] = engine_cls


def get_available_engines(pln: Union[Plan, dict[str]]) -> dict[str, Type[DoseEngineBase]]:
    """
    Get a list of available engines based on the plan.

    Parameters
    ----------
    pln : Union[Plan,dict]
        A Plan object.

    Returns
    -------
    list
        A list of available engines.
    """
    pln = validate_pln(pln)
    return {
        name: cls
        for name, cls in DOSE_ENGINES.items()
        if pln.radiation_mode in cls.possible_radiation_modes
    }


def get_engine(pln: Union[Plan, dict]) -> DoseEngineBase:
    """
    Factory function to get the appropriate engine based on the plan.

    Parameters
    ----------
    pln : Plan
        A Plan object.

    Returns
    -------
    Engine
        A Dose Engine object.
    """
    pln = validate_pln(pln)

    # Available engines
    engines = get_available_engines(pln)
    if len(engines) <= 0:
        raise ValueError(f"No engine available for radiation mode '{pln.radiation_mode}'.")

    engine_names = list(engines.keys())

    # Did the user provide an engine in the pln?
    if isinstance(pln.prop_dose_calc, DoseEngineBase):
        # The user provided an engine object, so lets use it
        # but warn the user if it is not in the available engines
        engine_name = pln.prop_dose_calc.short_name
        if engine_name not in engines:
            warnings.warn(f"Engine '{engine_name}' seems not to be valid for Plan setup.")
        return pln.prop_dose_calc

    if isinstance(pln.prop_dose_calc, dict):
        # The user provided a dictionary with engine parameters, so we need to find the engine name
        if "engine" in pln.prop_dose_calc:
            if pln.prop_dose_calc["engine"] in engines:
                return engines[pln.prop_dose_calc["engine"]](pln)
            warnings.warn(f"Engine '{pln.prop_dose_calc['engine']}' not available for Plan.")

        # If no engine name was found, we choose the first as default
        logger.warning(
            "No engine specified in Plan. Using first available engine %s.", engine_names[0]
        )
        return engines[engine_names[0]](pln)

    raise ValueError(f"No engine available for radiation mode '{pln.radiation_mode}'.")



from typing import Union

import numpy as np

from pyRadPlan.stf import validate_stf, SteeringInformation
from pyRadPlan.ct import validate_ct, CT
from pyRadPlan.cst import validate_cst, StructureSet
from pyRadPlan.plan import validate_pln, Plan
from pyRadPlan.dij import Dij

from .engines import get_engine


def calc_dose_influence(
    ct: Union[CT, dict],
    cst: Union[StructureSet, dict],
    stf: Union[SteeringInformation, dict],
    pln: Union[Plan, dict],
) -> Dij:
    """
    Calculate the dose influence matrix.

    Parameters
    ----------
    ct : CT
        A CT object.
    cst : CST
        A CST object.
    stf : STF
        A STF object.

    Returns
    -------
    Dij
        A Dij object.
    """

    ct = validate_ct(ct)
    cst = validate_cst(cst, ct=ct)
    stf = validate_stf(stf)
    pln = validate_pln(pln)

    engine = get_engine(pln)

    dij = engine.calc_dose_influence(ct, cst, stf)
    return dij


def calc_dose_forward(
    ct: Union[CT, dict],
    cst: Union[StructureSet, dict],
    stf: Union[SteeringInformation, dict],
    pln: Union[Plan, dict],
    weights: np.ndarray = None,
) -> Dij:
    """
    Calculate the dose forward matrix.

    Parameters
    ----------
    ct : CT
        A CT object.
    cst : CST
        A CST object.
    stf : STF
        A STF object.
    weights : np.ndarray
        The weights for the beamlets.

    Returns
    -------
    PlanResult
        A PlanResult object.
    """
    ct = validate_ct(ct)
    cst = validate_cst(cst, ct=ct)
    stf = validate_stf(stf)
    pln = validate_pln(pln)

    engine = get_engine(pln)

    dij = engine.calc_dose_forward(ct, cst, stf, weights)
    return dij



import pytest
from datetime import datetime
import numpy as np
from numpydantic import NDArray, Shape

from pyRadPlan.machines import (
    Machine,
    PhotonLINAC,
    IonAccelerator,
    load_from_name,
    IonPencilBeamKernel,
    PhotonSVDKernel,
)


def test_basic_machine_model():
    machine = Machine(
        radiation_mode="photons",
        description="A dummy machine",
        name="Generic",
        created_on="2021-01-01T00:00:00",
        last_modified="2021-01-01T00:00:00",
        created_by="Jay Doe",
        last_modified_by="Jay Doe",
        data_type="-",
        version="1.0.0",
    )

    assert isinstance(machine.created_on, datetime)
    assert isinstance(machine.last_modified, datetime)

    assert machine.radiation_mode == "photons"
    assert machine.description == "A dummy machine"
    assert machine.name == "Generic"
    assert machine.created_on == datetime.strptime("2021-01-01T00:00:00", "%Y-%m-%dT%H:%M:%S")
    assert machine.last_modified == datetime.strptime("2021-01-01T00:00:00", "%Y-%m-%dT%H:%M:%S")
    assert machine.created_by == "Jay Doe"
    assert machine.last_modified_by == "Jay Doe"
    assert machine.data_type == "-"
    assert machine.version == "1.0.0"


def test_machine_model_only_required():
    machine = Machine(radiation_mode="photons", name="Generic")
    assert machine.radiation_mode == "photons"
    assert machine.name == "Generic"
    assert machine.description == ""
    assert machine.created_on is None
    assert machine.last_modified is None
    assert machine.created_by == ""
    assert machine.last_modified_by == ""
    assert machine.data_type == "-"
    assert machine.version == "0.0.1"


def test_machine_model_invalid_version():
    with pytest.raises(ValueError):
        machine = Machine(radiation_mode="photons", name="Generic", version="1.0")


def test_machine_model_invalid_machine():
    with pytest.raises(ValueError):
        machine = Machine(radiation_mode="protons", name="")


def test_machine_model_missing_required():
    with pytest.raises(ValueError):
        machine = Machine()


def test_photon_machine_model_from_name():
    machine = load_from_name(radiation_mode="photons", machine_name="Generic")
    assert isinstance(machine, Machine)
    assert isinstance(machine, PhotonLINAC)
    assert machine.radiation_mode == "photons"
    assert machine.name == "Generic"


def test_photon_machine_model_svdpb_get_kernel():
    machine = load_from_name(radiation_mode="photons", machine_name="Generic")
    assert isinstance(machine, PhotonLINAC)
    kernel = machine.get_kernel_by_index(0)

    assert isinstance(kernel, PhotonSVDKernel)
    nk = kernel.num_kernel_components
    assert nk == 3
    nssds = kernel.kernel_ssds.shape[0]
    npos = kernel.kernel_pos.shape[0]
    assert isinstance(kernel.kernel_data, NDArray[Shape[f"{nssds}, {nk}, {npos}"], np.float64])

    kernel1 = kernel.get_kernels_at_ssd(750)
    assert isinstance(kernel1, NDArray[Shape[f"{nk}, {npos}"], np.float64])
    kernel2 = kernel.get_kernels_at_ssd(750.2)
    assert isinstance(kernel2, NDArray[Shape[f"{nk}, {npos}"], np.float64])
    assert np.isclose(kernel1, kernel2).all()

    # check ssd warnings
    with pytest.warns(UserWarning):
        kernel_low = kernel.get_kernels_at_ssd(250)
        assert isinstance(kernel_low, NDArray[Shape[f"{nk}, {npos}"], np.float64])
        assert np.isclose(kernel.get_kernels_at_ssd(kernel.kernel_ssds.min()), kernel_low).all()

    with pytest.warns(UserWarning):
        kernel_high = kernel.get_kernels_at_ssd(5000)
        assert isinstance(kernel_high, NDArray[Shape[f"{nk}, {npos}"], np.float64])
        assert np.isclose(kernel.get_kernels_at_ssd(kernel.kernel_ssds.max()), kernel_high).all()


# TODO: model dump and alias tests


def test_proton_machine_model_from_name():
    machine = load_from_name(radiation_mode="protons", machine_name="Generic")
    assert isinstance(machine, Machine)
    assert isinstance(machine, IonAccelerator)
    assert machine.radiation_mode == "protons"
    assert machine.name == "Generic"

    assert isinstance(machine.energies, np.ndarray)
    assert isinstance(machine.peak_positions, np.ndarray)
    assert machine.pb_kernels is not None
    assert isinstance(machine.sad, float)


def test_proton_machine_model_get_kernel():
    machine = load_from_name(radiation_mode="protons", machine_name="Generic")
    assert isinstance(machine, IonAccelerator)
    kernel = machine.get_kernel_by_index(0)
    assert isinstance(kernel, IonPencilBeamKernel)


def test_proton_machine_model_phase_space():
    machine = load_from_name(radiation_mode="protons", machine_name="Generic")
    assert isinstance(machine, IonAccelerator)
    for _, spectrum in machine.spectra.items():
        assert isinstance(spectrum.mean, float)
        assert isinstance(spectrum.sigma, float)
        assert spectrum.mean * spectrum.sigma / 100 == pytest.approx(spectrum.sigma_absolute)
        assert spectrum.sigma / 100 == pytest.approx(spectrum.sigma_relative)

    for _, focus_list in machine.foci.items():
        for focus in focus_list:
            assert isinstance(focus.emittance.sigma_x, float)
            assert isinstance(focus.emittance.sigma_y, float)
            assert isinstance(focus.emittance.div_x, float)
            assert isinstance(focus.emittance.div_y, float)
            assert isinstance(focus.emittance.corr_x, float)
            assert isinstance(focus.emittance.corr_y, float)


def test_carbon_machine_model_from_name():
    machine = load_from_name(radiation_mode="carbon", machine_name="Generic")
    assert isinstance(machine, Machine)
    assert isinstance(machine, IonAccelerator)
    assert machine.radiation_mode == "carbon"
    assert machine.name == "Generic"

    assert isinstance(machine.energies, np.ndarray)
    assert isinstance(machine.peak_positions, np.ndarray)
    assert machine.pb_kernels is not None
    assert isinstance(machine.sad, float)


def test_carbon_machine_model_get_kernel():
    machine = load_from_name(radiation_mode="carbon", machine_name="Generic")
    assert isinstance(machine, IonAccelerator)
    kernel = machine.get_kernel_by_index(0)
    assert isinstance(kernel, IonPencilBeamKernel)

    abratio = kernel.alpha_beta_ratio
    assert np.isclose(abratio, kernel.alpha_x / kernel.beta_x).all()

    assert kernel.alpha.shape == (abratio.size, kernel.depths.size)
    assert kernel.beta.shape == (abratio.size, kernel.depths.size)



import pytest
import numpy as np
import SimpleITK as sitk

# from pyRadPlan.core import Grid
from pyRadPlan.core.resample import resample_image, resample_numpy_array


@pytest.fixture
def sample_image():
    """Create a sample SimpleITK image for testing."""
    sample_array = np.random.rand(50, 100, 100) * 1000  # Random HU values
    image = sitk.GetImageFromArray(sample_array)
    image.SetOrigin((0, 0, 0))
    image.SetSpacing((1, 1, 2))
    image.SetDirection((1, 0, 0, 0, 1, 0, 0, 0, 1))
    return image


def test_resample_image2grid_with_spacing(sample_image):
    target_spacing = (3, 2, 2)
    resampled_image = resample_image(sample_image, target_grid_spacing=target_spacing)
    assert resampled_image.GetSpacing() == target_spacing
    assert resampled_image.GetSize() == (34, 50, 50)


def test_resample_image2grid_with_target_image(sample_image):
    target_image = sample_image
    resampled_image = resample_image(sample_image, target_image=target_image)
    assert resampled_image.GetSpacing() == target_image.GetSpacing()
    assert resampled_image.GetSize() == target_image.GetSize()


def test_resample_numpy_array(sample_image):
    sample_array = sitk.GetArrayFromImage(sample_image)
    resampled_array = resample_numpy_array(
        sample_array, sample_image, target_grid_spacing=(2, 2, 2)
    )
    assert resampled_array.shape == (50, 50, 50)


def test_resample_numpy_array_with_target_image(sample_image):
    sample_array = sitk.GetArrayFromImage(sample_image)
    resampled_array = resample_numpy_array(sample_array, sample_image, target_image=sample_image)
    assert resampled_array.shape == sample_array.shape



from ._matlab import MatRadEngineMatlab
from ._octave import MatRadEngineOctave
from ._engine import MatRadEngine

engine = None


def setEngine(engine_=None):
    global engine
    if isinstance(engine, MatRadEngine) or engine is None:
        engine = engine_
    else:
        raise ValueError("Invalid matRad engine")

    print("Engine set to " + engine.engineVersionStr + " using " + engine.matRadVersionStr)


__all__ = ["engine", "setEngine", "MatRadEngineMatlab", "MatRadEngineOctave"]



from typing import Union
import numpy as np
from numpy.typing import ArrayLike

from ...plan import Plan
from ._optiprob import NonLinearPlanningProblem
from ..solvers import NonLinearOptimizer


class SimpleLeastSquaresFluenceOptimization(NonLinearPlanningProblem):
    """Simple least squares fluence-based planning problem."""

    penalties: ArrayLike
    result: dict[str]

    def __init__(self, pln: Union[Plan, dict] = None):
        self.penalties = np.asarray([1000, 1])
        self.target_prescription = 2.0
        self._target_voxels = None
        self._patient_voxels = None

        self._grad_cache_intermediate = None
        self._grad_cache = None
        self._result_cache = None
        self._w_cache = None

        super().__init__(pln)

    def _initialize(self):
        super()._initialize()
        self._target_voxels = self._cst.target_union_voxels(order="numpy")
        self._patient_voxels = self._cst.patient_voxels(order="numpy")

        # Check if the solver is adequate to solve this problem
        # TODO: check that it can do constraints
        if not isinstance(self.solver, NonLinearOptimizer):
            raise ValueError("Solver must be an instance of SolverBase")

        self.solver.objective = self._objective_function
        self.solver.gradient = self._objective_gradient
        self.solver.bounds = (0.0, np.inf)
        self.solver.max_iter = 500
        self.solver.options = {
            "disp": True,
            "ftol": 1e-4,
        }

    def _objective_functions(self, x: np.ndarray) -> np.ndarray:
        """Define the objective functions."""

        if not np.array_equal(x, self._w_cache):
            self._result_cache = self._dij.get_result_arrays_from_intensity(x)
            self._w_cache = x.copy()

        dose = self._result_cache["physical_dose"]
        target_fun = (
            np.sum((dose[self._target_voxels] - self.target_prescription) ** 2)
            / self._target_voxels.size
        )
        patient_fun = np.sum((dose[self._patient_voxels]) ** 2) / self._patient_voxels.size
        return np.array([target_fun, patient_fun])

    def _objective_function(self, x: np.ndarray) -> np.float64:
        return np.dot(np.asarray(self.penalties), self._objective_functions(x))

    def _objective_jacobian(self, x: np.ndarray) -> np.ndarray:
        """Define the objective jacobian."""

        if not np.array_equal(x, self._w_cache):
            self._result_cache = self._dij.get_result_arrays_from_intensity(x)
            self._w_cache = x.copy()

        dose = self._result_cache["physical_dose"]

        if self._grad_cache_intermediate is None:
            self._grad_cache_intermediate = np.zeros((dose.size, 2))

        # We do no sanity checks here, i.e., the grids need to be the same

        self._grad_cache_intermediate[self._target_voxels, 0] = (
            2.0 / self._target_voxels.size * (dose[self._target_voxels] - self.target_prescription)
        )
        self._grad_cache_intermediate[self._patient_voxels, 1] = (
            2.0 / self._patient_voxels.size * (dose[self._patient_voxels])
        )

        self._grad_cache = self._dij.physical_dose.flat[0].T @ self._grad_cache_intermediate

        return self._grad_cache

    def _objective_gradient(self, x: np.ndarray) -> np.ndarray:
        return np.sum(self._objective_jacobian(x) * self.penalties, axis=1)

    def _objective_hessian(self, x: np.ndarray) -> np.ndarray:
        """Define the objective hessian."""
        return {}

    def _constraint_functions(self, x: np.ndarray) -> np.ndarray:
        """Define the constraint functions."""
        return None

    def _constraint_jacobian(self, x: np.ndarray) -> np.ndarray:
        """Define the constraint jacobian."""
        return None

    def _constraint_jacobian_structure(self) -> np.ndarray:
        """Define the constraint jacobian structure."""
        return None

    def _variable_bounds(self, x: np.ndarray) -> np.ndarray:
        """Define the variable bounds."""
        return {}

    def _solve(self) -> tuple[np.ndarray, dict]:
        """Solve the problem."""

        x0 = np.zeros((self._dij.total_num_of_bixels,), dtype=np.float64)
        return self.solver.solve(x0)



from typing import Optional, Annotated, Any
import warnings
from datetime import datetime
from pydantic import (
    Field,
    StringConstraints,
    field_validator,
    model_validator,
)
from numpydantic import NDArray, Shape
import numpy as np

from pyRadPlan.core import PyRadPlanBaseModel


class Machine(PyRadPlanBaseModel):
    """Base class for Machine objects
    Defines minimum meta-data a machine must hold:

    Attributes
    ----------
    radiation_mode : str
        The radiation mode of the machine.
    description : str
        The description of the machine.
    machine : str
        The name of the machine.
    """

    radiation_mode: str = Field()
    description: str = Field(default="")
    name: Annotated[str, StringConstraints(min_length=1)] = Field(alias="machine")

    created_on: Optional[datetime] = Field(default=None)
    last_modified: Optional[datetime] = Field(default=None)
    created_by: Optional[str] = Field(default="")
    last_modified_by: Optional[str] = Field(default="")
    data_type: Optional[str] = Field(default="-")
    version: Annotated[str, StringConstraints(pattern=r"^\d+\.\d+\.\d+$")] = Field(
        default="0.0.1", validate_default=True
    )

    @field_validator("created_on", "last_modified", mode="before")
    @classmethod
    def validate_datetime_variants(cls, v):
        # If it is a string, we try some additional formats in addition to
        # pydantics accepted datetime values
        # For example, matRad macines use the format "%d-%b-%Y" for some dates
        if isinstance(v, str):
            try_formats = ["%d-%b-%Y"]

            for fmt in try_formats:
                try:
                    return datetime.strptime(v, fmt)
                except ValueError:
                    pass

        return v


class ExternalBeamMachine(Machine):
    """
    Base class for Machine used for external irradation.

    Attributes
    ----------
    sad : float
        The source-to-axis distance of the machine
    energies : list[float]
        List of available photon energies
    """

    energies: NDArray[Shape["1-*"], np.float64]
    sad: float = Field(ge=0.0, description="Source-to-axis distance", alias="SAD")

    # Model sanitation / validation for compatibility
    @model_validator(mode="before")
    @classmethod
    def validate_machine_input(cls, data: Any) -> Any:
        """Validate the input data for the machine model before passing to
        pydantic.
        """
        if isinstance(data, dict):
            if "meta" in data and "data" in data:
                # looks like we loaded a matRad machine from a mat file

                # assign meta fields to base dict
                for key, value in data["meta"].items():
                    if key in data:
                        warnings.warn(
                            f"Meta data key {key} is already present in the data dictionary.",
                            UserWarning,
                        )

                    if key == "data":
                        warnings.warn("data key in metadata not allowed. Ignoring!")
                    data[key] = value

                # Clean up meta
                data.pop("meta")

                # prepare data structure for validation
                tabulated_energy_data = data.pop("data")
                cls._parse_tabulated_energy_data_from_mat(tabulated_energy_data, data)

        return data

    @classmethod
    def _parse_tabulated_energy_data_from_mat(
        cls, tabulated_energy_data: dict, returned_data: dict
    ):
        """Parse the tabulated energy data from a matRad machine file."""
        # prepare data structure for validation
        if isinstance(tabulated_energy_data, dict):
            # We might have a dictionary of lists
            if "energy" not in tabulated_energy_data:
                raise ValueError("Energy could not be found!")

            returned_data["energies"] = np.array(tabulated_energy_data["energy"], dtype=np.float64)

            if returned_data["energies"].size == 1:
                returned_data["energies"] = returned_data["energies"].flatten()

    def get_closest_energy_index(self, energy: float) -> int:
        """
        Given an energy value, return the closest matching index.

        Parameters
        ----------
        energy : float
            Energy value to search for

        Returns
        -------
        int
            Index of the energy level
        """
        return np.argmin(np.abs(self.energies - energy))

    def get_energy_index(self, energy: float, round_decimals=4) -> Optional[int]:
        """
        Given an energy value, return the index of the exact matching.

        Parameters
        ----------
        energy : float
            Energy value to search for
        round_decimals : int, optional
            Number of decimals to round the energy values to, by default 4

        Returns
        -------
        Optional[int]
            Index of the energy level
        """
        ix = np.where(np.round(self.energies, round_decimals) == np.round(energy, round_decimals))[
            0
        ]
        if len(ix) == 0:
            return None
        if len(ix) > 1:
            raise ValueError("Multiple matching energies found.")
        return ix[0]



from ._base import Machine, ExternalBeamMachine
from ._photons import PhotonLINAC, PhotonSVDKernel
from ._ions import IonAccelerator, IonPencilBeamKernel, LateralCutOff
from ._validate import validate_machine
from ._load import load_from_name, load_machine_from_mat, load_machine

__all__ = [
    "Machine",
    "ExternalBeamMachine",
    "PhotonLINAC",
    "PhotonSVDKernel",
    "load_from_name",
    "validate_machine",
    "IonAccelerator",
    "IonPencilBeamKernel",
    "LateralCutOff",
    "load_machine_from_mat",
    "load_machine",
]



"""Equivalent uniform dose objective."""

from typing import Annotated
from pydantic import Field

from numba import njit
from numpy import sum as npsum

from ._objective import Objective, ParameterMetadata


class EUD(Objective):
    """
    Equivalent uniform dose (EUD) objective.

    Attributes
    ----------
    k : float
        exponent.
    eud_ref : float
        reference value
    """

    name = "EUD"

    eud_ref: Annotated[float, Field(default=0.0, ge=0.0), ParameterMetadata(kind="reference")]
    k: Annotated[float, Field(default=1.0), ParameterMetadata()]

    def compute_objective(self, values):
        return _compute_objective(values, self.eud_ref, self.k)

    def compute_gradient(self, values):
        return _compute_gradient(values, self.eud_ref, self.k)


@njit
def _compute_objective(dose, eud_ref, eud_k):
    eud = (npsum(dose ** (1 / eud_k)) / len(dose)) ** eud_k

    return (eud - eud_ref) ** 2


@njit
def _compute_gradient(dose, eud_ref, eud_k):
    eud = (npsum(dose ** (1 / eud_k)) / len(dose)) ** eud_k
    eud_gradient = (
        npsum(dose ** (1 / eud_k)) ** (eud_k - 1) * dose ** (1 / eud_k - 1) / (len(dose) ** eud_k)
    )

    return 2.0 * (eud - eud_ref) * eud_gradient



import SimpleITK as sitk
import numpy as np

from pyRadPlan.dose import calc_dose_forward, calc_dose_influence
from pyRadPlan.dose.engines import (
    ParticleHongPencilBeamEngine,
    DoseEngineBase,
)


def test_ParticleHongPencilBeamEngine(test_data_protons):
    engine = ParticleHongPencilBeamEngine(test_data_protons[0])
    assert engine
    assert engine.name != None
    assert isinstance(engine, ParticleHongPencilBeamEngine)
    assert isinstance(engine, DoseEngineBase)


def test_protons_cd_forward(test_data_protons):
    pln, ct, cst, stf, dij, result = test_data_protons

    result_py = calc_dose_forward(ct, cst, stf, pln, weights=None)
    result_py = sitk.GetArrayFromImage(result_py["physical_dose"])

    result_matRad_rot = np.swapaxes(result["physicalDose"], 0, 1)
    # only comparing to 1e-4 since matRad rounds to 4 digits
    assert np.allclose(result_py, result_matRad_rot, atol=1e-4)

    # plot_slice(
    #     ct=ct,
    #     cst=cst,
    #     overlay=result_py-result_matRad_rot,
    #     view_slice=5,
    #     plane="axial",
    #     overlay_unit="Gy",
    #     plt_show = True,
    #     use_global_max = False,
    # )


def test_helium_cd_forward(test_data_helium):
    pln, ct, cst, stf, dij, result = test_data_helium

    result_py = calc_dose_forward(ct, cst, stf, pln, weights=None)
    result_py = sitk.GetArrayFromImage(result_py["physical_dose"])

    result_matRad_rot = np.swapaxes(result["physicalDose"], 0, 1)
    # only comparing to 1e-4 since matRad rounds to 4 digits
    assert np.allclose(result_py, result_matRad_rot, atol=1e-4)


def test_carbon_cd_forward(test_data_carbon):
    pln, ct, cst, stf, dij, result = test_data_carbon

    result_py = calc_dose_forward(ct, cst, stf, pln)
    result_py = sitk.GetArrayFromImage(result_py["physical_dose"])

    result_matRad_rot = np.swapaxes(result["physicalDose"], 0, 1)
    # only comparing to 1e-4 since matRad rounds to 4 digits
    assert np.allclose(result_py, result_matRad_rot, atol=1e-4)


def test_protons_cd_influence(test_data_protons):
    pln, ct, cst, stf, dij, result = test_data_protons

    dij_py = calc_dose_influence(ct, cst, stf, pln)

    physical_dose_py_dense = dij_py.physical_dose.flat[0].toarray()
    physical_dose_mat_dense = dij.physical_dose.flat[0].toarray()

    assert np.allclose(physical_dose_py_dense, physical_dose_mat_dense, atol=1e-6)


def test_helium_cd_influence(test_data_helium):
    pln, ct, cst, stf, dij, result = test_data_helium

    dij_py = calc_dose_influence(ct, cst, stf, pln)

    physical_dose_py_dense = dij_py.physical_dose.flat[0].toarray()
    physical_dose_mat_dense = dij.physical_dose.flat[0].toarray()

    assert np.allclose(physical_dose_py_dense, physical_dose_mat_dense, atol=1e-6)


def test_carbon_cd_influence(test_data_carbon):
    pln, ct, cst, stf, dij, result = test_data_carbon

    dij_py = calc_dose_influence(ct, cst, stf, pln)

    physical_dose_py_dense = dij_py.physical_dose.flat[0].toarray()
    physical_dose_mat_dense = dij.physical_dose.flat[0].toarray()

    assert np.allclose(physical_dose_py_dense, physical_dose_mat_dense, atol=1e-6)
